import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
from xorq.tests.util import assert_frame_equal


def test_simple_agg_ops_read_parquet(data_dir):
    path = data_dir / "parquet" / "functional_alltypes.parquet"

    key_big_int_col = "bigint_col"
    con = xo.connect()
    t = con.read_parquet(path)
    result = (
        t.group_by(key_big_int_col)
        .aggregate(mean=t.double_col.mean(), max=t.timestamp_col.max())
        .execute()
    )
    assert result is not None


def test_simple_agg_ops_read_csv(data_dir):
    path = data_dir / "csv" / "functional_alltypes.csv"

    key_big_int_col = "bigint_col"
    con = xo.connect()
    t = con.read_csv(path)
    result = (
        t.group_by(key_big_int_col)
        .aggregate(mean=t.double_col.mean(), max=t.timestamp_col.max())
        .execute()
    )
    assert result is not None


def test_memtable_ops_dict():
    t = xo.memtable({"s": ["aaa", "a", "aa"]})
    assert t.s.length().execute().gt(0).all()


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(
            lambda: xo.memtable([(1, 2.0, "3")], columns=list("abc")),
            pd.DataFrame([(1, 2.0, "3")], columns=list("abc")),
            id="simple",
        ),
        param(
            lambda: xo.memtable([(1, 2.0, "3")]),
            pd.DataFrame([(1, 2.0, "3")], columns=["col0", "col1", "col2"]),
            id="simple_auto_named",
        ),
        param(
            lambda: xo.memtable(
                pd.DataFrame({"a": [1], "b": [2.0], "c": ["3"]}).astype(
                    {"a": "int8", "b": "float32"}
                )
            ),
            pd.DataFrame([(1, 2.0, "3")], columns=list("abc")).astype(
                {"a": "int8", "b": "float32"}
            ),
            id="dataframe",
        ),
        param(
            lambda: xo.memtable([{"a": 1}, {"a": 2}]),
            pd.DataFrame({"a": [1, 2]}),
            id="list_of_dicts",
        ),
    ],
)
def test_in_memory_table(expr, expected):
    result = expr().execute()
    assert_frame_equal(result, expected)



from __future__ import annotations

from xorq.vendor.ibis.formats.pandas import PandasData


class DuckDBPandasData(PandasData):
    @staticmethod
    def convert_Array(s, dtype, pandas_type):
        return s.replace(float("nan"), None)



from abc import (
    ABC,
    abstractmethod,
)

import pyarrow as pa
import pyarrow.flight as paf
import toolz
from cloudpickle import dumps, loads

from xorq.common.utils import classproperty


make_flight_result = toolz.compose(
    paf.Result,
    pa.py_buffer,
    dumps,
)


class AbstractAction(ABC):
    @classproperty
    @abstractmethod
    def name(cls):
        pass

    @classproperty
    @abstractmethod
    def description(cls):
        pass

    @classproperty
    @abstractmethod
    def do_action(cls, server, context, action):
        pass


class HealthCheckAction(AbstractAction):
    @classproperty
    def name(cls):
        return "healthcheck"

    @classproperty
    def description(cls):
        return "NOP: check that communication is established"

    @classmethod
    def do_action(cls, server, context, action):
        yield make_flight_result(None)


class ListActionsAction(AbstractAction):
    @classproperty
    def name(cls):
        return "list-actions"

    @classproperty
    def description(cls):
        return "Get a list of all actions available on this server."

    @classmethod
    def do_action(cls, server, context, action):
        yield make_flight_result(
            tuple(action.name for action in server.actions.values())
        )


class ListExchangesAction(AbstractAction):
    @classproperty
    def name(cls):
        return "list-exchanges"

    @classproperty
    def description(cls):
        return "Get a list of all exchange commands available on this server."

    @classmethod
    def do_action(cls, server, context, action):
        yield make_flight_result(
            tuple(exchanger_command for exchanger_command in server.exchangers.keys())
        )


class AddActionAction(AbstractAction):
    @classproperty
    def name(cls):
        return "add-action"

    @classproperty
    def description(cls):
        return "Add an action to the server's repertoire of actions"

    @classmethod
    def do_action(cls, server, context, action):
        action_class = loads(action.body)
        server.actions[action_class.name] = action_class
        yield make_flight_result(None)


class AddExchangeAction(AbstractAction):
    @classproperty
    def name(cls):
        return "add-exchange"

    @classproperty
    def description(cls):
        return "Add an exchange to the server's repertoire of exchanges"

    @classmethod
    def do_action(cls, server, context, action):
        exchange_class = loads(action.body)
        # HACK: stopgap until alias / command is specifiable via do-action
        server.exchangers[exchange_class.command] = exchange_class
        server.exchangers["default"] = exchange_class
        yield make_flight_result(None)


class QueryExchangeAction(AbstractAction):
    @classproperty
    def name(cls):
        return "query-exchange"

    @classproperty
    def description(cls):
        return "Get metadata about a particular exchange available on this server."

    @classmethod
    def do_action(cls, server, context, action):
        exchange_name = loads(action.body)
        exchanger = server.exchangers.get(exchange_name)
        query_result = exchanger.query_result if exchanger else None
        yield make_flight_result(query_result)


class GetExchangeAction(AbstractAction):
    @classproperty
    def name(cls):
        return "get-exchange"

    @classproperty
    def description(cls):
        return "Get a particular exchange available on this server."

    @classmethod
    def do_action(cls, server, context, action):
        exchange_name = loads(action.body)
        exchanger = server.exchangers.get(exchange_name)
        yield make_flight_result(exchanger)


class ListTablesAction(AbstractAction):
    @classproperty
    def name(cls):
        return "list_tables"

    @classproperty
    def description(cls):
        return "Get the names of all tables available on this server."

    @classmethod
    def do_action(cls, server, context, action):
        kwargs = loads(action.body)
        tables = server._conn.list_tables(**kwargs)
        yield make_flight_result(tables)


class TableInfoAction(AbstractAction):
    @classproperty
    def name(cls):
        return "table_info"

    @classproperty
    def description(cls):
        return "Get info about a particular table available on this server."

    @classmethod
    def do_action(cls, server, context, action):
        kwargs = loads(action.body)
        table_name = kwargs.pop("table_name")
        schema = server._conn.get_schema(table_name, **kwargs)
        yield make_flight_result(schema)


class DropTableAction(AbstractAction):
    @classproperty
    def name(cls):
        return "drop_table"

    @classproperty
    def description(cls):
        return "Drop a table on this server."

    @classmethod
    def do_action(cls, server, context, action):
        kwargs = loads(action.body)
        table_name = kwargs.pop("name")
        server._conn.drop_table(table_name, **kwargs)
        yield make_flight_result(f"dropped table {table_name}")


class DropViewAction(AbstractAction):
    @classproperty
    def name(cls):
        return "drop_view"

    @classproperty
    def description(cls):
        return "Drop a view on this server."

    @classmethod
    def do_action(cls, server, context, action):
        kwargs = loads(action.body)
        table_name = kwargs.pop("name")
        server._conn.drop_view(table_name, **kwargs)
        yield make_flight_result(f"dropped view {table_name}")


class ReadParquetAction(AbstractAction):
    @classproperty
    def name(cls):
        return "read_parquet"

    @classproperty
    def description(cls):
        return "Read parquet files into this server."

    @classmethod
    def do_action(cls, server, context, action):
        args = loads(action.body)

        table_name = args["table_name"]
        source_list = args["source_list"]
        table = server._conn.read_parquet(source_list, table_name=table_name)
        yield make_flight_result(table.get_name())


class VersionAction(AbstractAction):
    @classproperty
    def name(cls):
        return "version"

    @classproperty
    def description(cls):
        return "Return the version of the underlying backend"

    @classmethod
    def do_action(cls, server, context, action):
        yield make_flight_result(server._conn.version)


class GetSchemaQueryAction(AbstractAction):
    @classproperty
    def name(cls):
        return "get_schema_using_query"

    @classmethod
    def description(cls):
        return "Get the schema of query result"

    @classmethod
    def do_action(cls, server, context, action):
        query = loads(action.body)
        schema = server._conn._get_schema_using_query(query)
        yield make_flight_result(schema)


actions = {
    action.name: action
    for action in (
        HealthCheckAction,
        ListActionsAction,
        ListExchangesAction,
        QueryExchangeAction,
        AddActionAction,
        AddExchangeAction,
        ListTablesAction,
        TableInfoAction,
        DropTableAction,
        DropViewAction,
        ReadParquetAction,
        VersionAction,
        GetSchemaQueryAction,
        GetExchangeAction,
    )
}



from __future__ import annotations

import math
import operator
from functools import partial, reduce

import sqlglot as sg
import sqlglot.expressions as sge
import toolz

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis.backends.sql.compilers.base import (
    FALSE,
    NULL,
    STAR,
    TRUE,
    AggGen,
    SQLGlotCompiler,
)
from xorq.vendor.ibis.backends.sql.datatypes import TrinoType
from xorq.vendor.ibis.backends.sql.dialects import Trino
from xorq.vendor.ibis.backends.sql.rewrites import (
    exclude_unsupported_window_frame_from_ops,
    split_select_distinct_with_order_by,
)
from xorq.vendor.ibis.util import gen_name


class TrinoCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = Trino
    type_mapper = TrinoType

    agg = AggGen(supports_filter=True, supports_order_by=True)

    rewrites = (
        exclude_unsupported_window_frame_from_ops,
        *SQLGlotCompiler.rewrites,
    )
    post_rewrites = (split_select_distinct_with_order_by,)
    quoted = True

    NAN = sg.func("nan")
    POS_INF = sg.func("infinity")
    NEG_INF = -POS_INF

    UNSUPPORTED_OPS = (
        ops.Median,
        ops.RowID,
        ops.TimestampBucket,
    )

    LOWERED_OPS = {
        ops.Sample: None,
    }

    SIMPLE_OPS = {
        ops.Arbitrary: "any_value",
        ops.Pi: "pi",
        ops.E: "e",
        ops.RegexReplace: "regexp_replace",
        ops.Map: "map",
        ops.MapKeys: "map_keys",
        ops.MapLength: "cardinality",
        ops.MapMerge: "map_concat",
        ops.MapValues: "map_values",
        ops.Log2: "log2",
        ops.Log10: "log10",
        ops.IsNan: "is_nan",
        ops.IsInf: "is_infinite",
        ops.StringToTimestamp: "date_parse",
        ops.Strftime: "date_format",
        ops.ExtractEpochSeconds: "to_unixtime",
        ops.ExtractWeekOfYear: "week_of_year",
        ops.ExtractDayOfYear: "day_of_year",
        ops.ExtractMillisecond: "millisecond",
        ops.ArrayUnion: "array_union",
        ops.ArrayRemove: "array_remove",
        ops.ArrayFlatten: "flatten",
        ops.ArraySort: "array_sort",
        ops.ArrayDistinct: "array_distinct",
        ops.ArrayLength: "cardinality",
        ops.ArrayIntersect: "array_intersect",
        ops.BitAnd: "bitwise_and_agg",
        ops.BitOr: "bitwise_or_agg",
        ops.TypeOf: "typeof",
        ops.Levenshtein: "levenshtein_distance",
        ops.ExtractProtocol: "url_extract_protocol",
        ops.ExtractHost: "url_extract_host",
        ops.ExtractPath: "url_extract_path",
        ops.ExtractFragment: "url_extract_fragment",
        ops.ArrayPosition: "array_position",
        ops.ExtractIsoYear: "year_of_week",
    }

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def visit_Sample(
        self, op, *, parent, fraction: float, method: str, seed: int | None, **_
    ):
        if seed is not None:
            raise com.UnsupportedOperationError(
                "`Table.sample` with a random seed is unsupported"
            )
        sample = sge.TableSample(
            method="bernoulli" if method == "row" else "system",
            percent=sge.convert(fraction * 100.0),
            seed=None if seed is None else sge.convert(seed),
        )
        return self._make_sample_backwards_compatible(sample=sample, parent=parent)

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                "Trino does not support `sample` correlation"
            )
        if (left_type := op.left.dtype).is_boolean():
            left = self.cast(left, dt.Int32(nullable=left_type.nullable))

        if (right_type := op.right.dtype).is_boolean():
            right = self.cast(right, dt.Int32(nullable=right_type.nullable))

        return self.agg.corr(left, right, where=where)

    def visit_ApproxQuantile(self, op, *, arg, quantile, where):
        if not op.arg.dtype.is_floating():
            arg = self.cast(arg, dt.float64)
        return self.agg.approx_quantile(arg, quantile, where=where)

    visit_ApproxMultiQuantile = visit_ApproxQuantile

    def visit_BitXor(self, op, *, arg, where):
        a, b = map(sg.to_identifier, "ab")
        input_fn = combine_fn = sge.Lambda(
            this=sge.BitwiseXor(this=a, expression=b), expressions=[a, b]
        )
        return self.agg.reduce_agg(arg, 0, input_fn, combine_fn, where=where)

    def visit_ArrayRepeat(self, op, *, arg, times):
        return self.f.flatten(self.f.repeat(arg, times))

    def visit_ArraySlice(self, op, *, arg, start, stop):
        def _neg_idx_to_pos(n, idx):
            return self.if_(idx < 0, n + self.f.greatest(idx, -n), idx)

        arg_length = self.f.cardinality(arg)

        if start is None:
            start = 0
        else:
            start = self.f.least(arg_length, _neg_idx_to_pos(arg_length, start))

        if stop is None:
            stop = arg_length
        else:
            stop = _neg_idx_to_pos(arg_length, stop)

        return self.f.slice(arg, start + 1, stop - start)

    def visit_ArrayMap(self, op, *, arg, param, body):
        return self.f.transform(arg, sge.Lambda(this=body, expressions=[param]))

    def visit_ArrayFilter(self, op, *, arg, param, body):
        return self.f.filter(arg, sge.Lambda(this=body, expressions=[param]))

    def visit_ArrayContains(self, op, *, arg, other):
        return self.if_(
            arg.is_(sg.not_(NULL)),
            self.f.coalesce(self.f.contains(arg, other), FALSE),
            NULL,
        )

    def visit_ArrayCollect(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.array_agg(arg, where=where, order_by=order_by)

    def visit_JSONGetItem(self, op, *, arg, index):
        fmt = "%d" if op.index.dtype.is_integer() else '"%s"'
        return self.f.json_extract(arg, self.f.format(f"$[{fmt}]", index))

    def visit_UnwrapJSONString(self, op, *, arg):
        return self.f.json_value(
            self.f.json_format(arg), 'strict $?($.type() == "string")'
        )

    def visit_UnwrapJSONInt64(self, op, *, arg):
        value = self.f.json_value(
            self.f.json_format(arg), 'strict $?($.type() == "number")'
        )
        return self.cast(
            self.if_(self.f.regexp_like(value, r"^\d+$"), value, NULL), op.dtype
        )

    def visit_UnwrapJSONFloat64(self, op, *, arg):
        return self.cast(
            self.f.json_value(
                self.f.json_format(arg), 'strict $?($.type() == "number")'
            ),
            op.dtype,
        )

    def visit_UnwrapJSONBoolean(self, op, *, arg):
        return self.cast(
            self.f.json_value(
                self.f.json_format(arg), 'strict $?($.type() == "boolean")'
            ),
            op.dtype,
        )

    def visit_DayOfWeekIndex(self, op, *, arg):
        return self.cast(sge.paren(self.f.anon.dow(arg) + 6, copy=False) % 7, op.dtype)

    def visit_DayOfWeekName(self, op, *, arg):
        return self.f.date_format(arg, "%W")

    def visit_StrRight(self, op, *, arg, nchars):
        return self.f.substr(arg, -nchars)

    def visit_EndsWith(self, op, *, arg, end):
        return self.f.substr(arg, -self.f.length(end)).eq(end)

    def visit_Repeat(self, op, *, arg, times):
        return self.f.array_join(self.f.repeat(arg, times), "")

    def visit_DateTimestampTruncate(self, op, *, arg, unit):
        _truncate_precisions = {
            # ms unit is not yet officially documented but it works
            "ms": "millisecond",
            "s": "second",
            "m": "minute",
            "h": "hour",
            "D": "day",
            "W": "week",
            "M": "month",
            "Q": "quarter",
            "Y": "year",
        }

        if (precision := _truncate_precisions.get(unit.short)) is None:
            raise com.UnsupportedOperationError(
                f"Unsupported truncate unit {op.unit!r}"
            )
        return self.f.date_trunc(precision, arg)

    visit_DateTruncate = visit_TimestampTruncate = visit_DateTimestampTruncate

    def visit_DateFromYMD(self, op, *, year, month, day):
        return self.f.from_iso8601_date(
            self.f.format("%04d-%02d-%02d", year, month, day)
        )

    def visit_TimeFromHMS(self, op, *, hours, minutes, seconds):
        return self.cast(
            self.f.format("%02d:%02d:%02d", hours, minutes, seconds), dt.time
        )

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds
    ):
        return self.cast(
            self.f.from_iso8601_timestamp(
                self.f.format(
                    "%04d-%02d-%02dT%02d:%02d:%02d",
                    year,
                    month,
                    day,
                    hours,
                    minutes,
                    seconds,
                )
            ),
            dt.timestamp,
        )

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        short = unit.short
        if short == "ms":
            res = self.f.from_unixtime(self.f.floor(arg / 1_000))
        elif short == "s":
            res = self.f.from_unixtime(arg)
        elif short == "us":
            res = self.f.from_unixtime_nanos((arg - arg % 1_000_000) * 1_000)
        elif short == "ns":
            res = self.f.from_unixtime_nanos(arg - arg % 1_000_000_000)
        else:
            raise com.UnsupportedOperationError(f"{unit!r} unit is not supported")
        return self.cast(res, op.dtype)

    def visit_InSubquery(self, op, *, rel, needle):
        # cast the needle to the same type as the column being queried, since
        # trino is very strict about structs
        if op.needle.dtype.is_struct():
            needle = self.cast(
                sge.Struct.from_arg_list([needle]), op.rel.schema.as_struct()
            )

        return super().visit_InSubquery(op, rel=rel, needle=needle)

    def visit_StructColumn(self, op, *, names, values):
        return self.cast(sge.Struct(expressions=list(values)), op.dtype)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_floating():
            if math.isfinite(value):
                return self.cast(value, dtype)
            return super().visit_NonNullLiteral(op, value=value, dtype=dtype)
        elif dtype.is_struct():
            items = [
                self.visit_Literal(ops.Literal(v, fdtype), value=v, dtype=fdtype)
                for fdtype, v in zip(dtype.types, value.values())
            ]
            return self.cast(sge.Struct(expressions=items), dtype)
        elif dtype.is_timestamp():
            return self.cast(self.f.from_iso8601_timestamp(value.isoformat()), dtype)
        elif dtype.is_date():
            return self.f.from_iso8601_date(value.isoformat())
        elif dtype.is_time():
            return self.cast(value.isoformat(), dtype)
        elif dtype.is_interval():
            return self._make_interval(sge.convert(str(value)), dtype.unit)
        elif dtype.is_binary():
            return self.f.from_hex(value.hex())
        else:
            return None

    def visit_Log(self, op, *, arg, base):
        return self.f.log(base, arg, dialect=self.dialect)

    def visit_MapGet(self, op, *, arg, key, default):
        return self.f.coalesce(self.f.element_at(arg, key), default)

    def visit_MapContains(self, op, *, arg, key):
        return self.f.contains(self.f.map_keys(arg), key)

    def visit_ExtractFile(self, op, *, arg):
        return self.f.concat_ws(
            "?",
            self.f.nullif(self.f.url_extract_path(arg), ""),
            self.f.nullif(self.f.url_extract_query(arg), ""),
        )

    def visit_ExtractQuery(self, op, *, arg, key):
        if key is None:
            return self.f.url_extract_query(arg)
        return self.f.url_extract_parameter(arg, key)

    def visit_Cot(self, op, *, arg):
        return 1.0 / self.f.tan(arg)

    def visit_StringAscii(self, op, *, arg):
        return self.f.codepoint(
            sge.Cast(
                this=self.f.substr(arg, 1, 2),
                to=sge.DataType(
                    this=sge.DataType.Type.VARCHAR,
                    expressions=[sge.DataTypeParam(this=sge.convert(1))],
                ),
            )
        )

    def visit_ArrayStringJoin(self, op, *, sep, arg):
        return self.f.array_join(arg, sep)

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.f.element_at(
            self.agg.array_agg(arg, where=where, order_by=order_by), 1
        )

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.f.element_at(
            self.agg.array_agg(arg, where=where, order_by=order_by), -1
        )

    def visit_GroupConcat(self, op, *, arg, sep, where, order_by):
        cond = arg.is_(sg.not_(NULL, copy=False))
        where = cond if where is None else sge.And(this=cond, expression=where)
        array = self.agg.array_agg(
            self.cast(arg, dt.string), where=where, order_by=order_by
        )
        return self.f.array_join(array, sep)

    def visit_ArrayZip(self, op, *, arg):
        max_zip_arguments = 5
        chunks = (
            (len(chunk), self.f.zip(*chunk) if len(chunk) > 1 else chunk[0])
            for chunk in toolz.partition_all(max_zip_arguments, arg)
        )

        def combine_zipped(left, right):
            left_n, left_chunk = left
            x, y = map(sg.to_identifier, "xy")

            lhs = list(map(x.__getitem__, range(left_n))) if left_n > 1 else [x]

            right_n, right_chunk = right
            rhs = list(map(y.__getitem__, range(right_n))) if right_n > 1 else [y]

            zipped_chunk = self.f.zip_with(
                left_chunk,
                right_chunk,
                sge.Lambda(this=self.f.row(*lhs, *rhs), expressions=[x, y]),
            )
            return left_n + right_n, zipped_chunk

        all_n, chunk = reduce(combine_zipped, chunks)
        assert all_n == len(op.dtype.value_type)
        return chunk

    def visit_ExtractMicrosecond(self, op, *, arg):
        # trino only seems to store milliseconds, but the result of formatting
        # always pads the right with 000
        return self.cast(self.f.date_format(arg, "%f"), dt.int32)

    def visit_TemporalDelta(self, op, *, part, left, right):
        # trino truncates _after_ the delta, whereas many other backends
        # truncate each operand
        dialect = self.dialect
        return self.f.date_diff(
            part,
            self.f.date_trunc(part, right, dialect=dialect),
            self.f.date_trunc(part, left, dialect=dialect),
            dialect=dialect,
        )

    visit_TimeDelta = visit_DateDelta = visit_TimestampDelta = visit_TemporalDelta

    def _make_interval(self, arg, unit):
        short = unit.short
        if short in ("Q", "W"):
            raise com.UnsupportedOperationError(f"Interval unit {unit!r} not supported")

        if isinstance(arg, sge.Literal):
            # force strings in interval literals because trino requires it
            arg.args["is_string"] = True
            return super()._make_interval(arg, unit)

        elif short in ("Y", "M"):
            return arg * super()._make_interval(sge.convert("1"), unit)
        elif short in ("D", "h", "m", "s", "ms", "us"):
            return self.f.parse_duration(
                self.f.concat(self.cast(arg, dt.string), short.lower())
            )
        else:
            raise com.UnsupportedOperationError(
                f"Interval unit {unit.name!r} not supported"
            )

    def visit_Range(self, op, *, start, stop, step):
        def zero_value(dtype):
            if dtype.is_interval():
                # the unit doesn't matter here, because e.g. 0d = 0s
                return self.f.parse_duration("0s")
            return 0

        def interval_sign(v):
            zero = self.f.parse_duration("0s")
            return sge.Case(
                ifs=[
                    self.if_(v.eq(zero), 0),
                    self.if_(v < zero, -1),
                    self.if_(v > zero, 1),
                ]
            )

        def _sign(value, dtype):
            if dtype.is_interval():
                return interval_sign(value)
            return self.f.sign(value)

        step_dtype = op.step.dtype
        zero = zero_value(step_dtype)
        return self.if_(
            sg.and_(
                self.f.nullif(step, zero).is_(sg.not_(NULL)),
                _sign(step, step_dtype).eq(_sign(stop - start, step_dtype)),
            ),
            self.f.array_remove(self.f.sequence(start, stop, step), stop),
            self.f.array(),
        )

    visit_IntegerRange = visit_TimestampRange = visit_Range

    def visit_ArrayIndex(self, op, *, arg, index):
        return self.f.element_at(arg, index + 1)

    def visit_Cast(self, op, *, arg, to):
        from_ = op.arg.dtype
        if from_.is_integer() and to.is_timestamp():
            return self.f.from_unixtime(arg, to.timezone or "UTC")
        return super().visit_Cast(op, arg=arg, to=to)

    def visit_CountDistinctStar(self, op, *, arg, where):
        make_col = partial(sg.column, table=arg.alias_or_name, quoted=self.quoted)
        row = self.f.row(*map(make_col, op.arg.schema.names))
        return self.agg.count(sge.Distinct(expressions=[row]), where=where)

    def visit_ArrayConcat(self, op, *, arg):
        return self.f.concat(*arg)

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.strpos(haystack, needle) > 0

    def visit_RegexpExtract(self, op, *, arg, pattern, index):
        # sqlglot doesn't support the third `group` argument for trino so work
        # around that limitation using an anonymous function
        return self.f.anon.regexp_extract(arg, pattern, index)

    def visit_ToJSONMap(self, op, *, arg):
        return self.cast(
            self.f.json_parse(
                self.f.json_query(
                    self.f.json_format(arg), 'strict $?($.type() == "object")'
                )
            ),
            dt.Map(dt.string, dt.json),
        )

    def visit_ToJSONArray(self, op, *, arg):
        return self.cast(
            self.f.json_parse(
                self.f.json_query(
                    self.f.json_format(arg), 'strict $?($.type() == "array")'
                )
            ),
            dt.Array(dt.json),
        )

    def visit_TableUnnest(
        self,
        op,
        *,
        parent,
        column,
        column_name: str,
        offset: str | None,
        keep_empty: bool,
    ):
        quoted = self.quoted

        column_alias = sg.to_identifier(gen_name("table_unnest_column"), quoted=quoted)

        parent_schema = op.parent.schema
        overlaps_with_parent = column_name in parent_schema
        computed_column = column_alias.as_(column_name, quoted=quoted)

        parent_alias_or_name = parent.alias_or_name

        selcols = []

        if overlaps_with_parent:
            column_alias_or_name = column.alias_or_name
            selcols.extend(
                sg.column(col, table=parent_alias_or_name, quoted=quoted)
                if col != column_alias_or_name
                else computed_column
                for col in parent_schema.names
            )
        else:
            selcols.append(
                sge.Column(
                    this=STAR,
                    table=sg.to_identifier(parent_alias_or_name, quoted=quoted),
                )
            )
            selcols.append(computed_column)

        if offset is not None:
            offset_name = offset
            offset = sg.to_identifier(offset_name, quoted=quoted)
            selcols.append((offset - 1).as_(offset_name, quoted=quoted))

        unnest = sge.Unnest(
            expressions=[column],
            alias=sge.TableAlias(
                this=sg.to_identifier(gen_name("table_unnest"), quoted=quoted),
                columns=[column_alias],
            ),
            offset=offset,
        )
        return (
            sg.select(*selcols)
            .from_(parent)
            .join(
                unnest,
                on=None if not keep_empty else TRUE,
                join_type="CROSS" if not keep_empty else "LEFT",
            )
        )

    def visit_ArrayAny(self, op, *, arg):
        x = sg.to_identifier("x", quoted=self.quoted)
        identity = sge.Lambda(this=x, expressions=[x])
        is_not_null = sge.Lambda(this=x.is_(sg.not_(NULL)), expressions=[x])
        return self.f.any_match(
            self.f.nullif(self.f.filter(arg, is_not_null), self.f.array()), identity
        )

    def visit_ArrayAll(self, op, *, arg):
        x = sg.to_identifier("x", quoted=self.quoted)
        identity = sge.Lambda(this=x, expressions=[x])
        is_not_null = sge.Lambda(this=x.is_(sg.not_(NULL)), expressions=[x])
        return self.f.all_match(
            self.f.nullif(self.f.filter(arg, is_not_null), self.f.array()), identity
        )

    def visit_ArrayMin(self, op, *, arg):
        x = sg.to_identifier("x", quoted=self.quoted)
        func = sge.Lambda(this=x.is_(sg.not_(NULL)), expressions=[x])
        return self.f.array_min(self.f.filter(arg, func))

    def visit_ArrayMax(self, op, *, arg):
        x = sg.to_identifier("x", quoted=self.quoted)
        func = sge.Lambda(this=x.is_(sg.not_(NULL)), expressions=[x])
        return self.f.array_max(self.f.filter(arg, func))

    def visit_ArraySumAgg(self, op, *, arg, output):
        quoted = self.quoted
        dot = lambda a, f: sge.Dot.build((a, sge.to_identifier(f, quoted=quoted)))  # noqa: E731
        state_dtype = dt.Struct({"sum": op.dtype, "count": dt.int64})
        initial_state = self.cast(
            sge.Struct.from_arg_list([sge.convert(0), sge.convert(0)]), state_dtype
        )

        s = sg.to_identifier("s", quoted=quoted)
        x = sg.to_identifier("x", quoted=quoted)

        s_sum = dot(s, "sum")
        s_count = dot(s, "count")

        input_fn_body = self.cast(
            sge.Struct.from_arg_list(
                [
                    x + self.f.coalesce(s_sum, 0),
                    s_count + self.if_(x.is_(sg.not_(NULL)), 1, 0),
                ]
            ),
            state_dtype,
        )
        input_fn = sge.Lambda(this=input_fn_body, expressions=[s, x])

        output_fn_body = self.if_(s_count > 0, output(s_sum, s_count), NULL)
        return self.f.reduce(
            arg,
            initial_state,
            input_fn,
            sge.Lambda(this=output_fn_body, expressions=[s]),
        )

    def visit_ArraySum(self, op, *, arg):
        return self.visit_ArraySumAgg(op, arg=arg, output=lambda sum, _: sum)

    def visit_ArrayMean(self, op, *, arg):
        return self.visit_ArraySumAgg(op, arg=arg, output=operator.truediv)


compiler = TrinoCompiler()



from __future__ import annotations

import math
import numbers
from abc import abstractmethod
from collections.abc import Callable, Mapping, Sequence
from enum import Enum
from inspect import Parameter
from typing import (
    Annotated,
    ForwardRef,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from typing import Any as AnyType

import toolz
from typing_extensions import GenericMeta

from xorq.vendor.ibis.common.bases import FrozenSlotted as Slotted
from xorq.vendor.ibis.common.bases import Hashable, Singleton
from xorq.vendor.ibis.common.collections import (
    FrozenDict,
    RewindableIterator,
    frozendict,
)
from xorq.vendor.ibis.common.deferred import (
    Deferred,
    Factory,
    Resolver,
    Variable,
    _,  # noqa: F401
    resolver,
)
from xorq.vendor.ibis.common.typing import (
    Coercible,
    CoercionError,
    Sentinel,
    UnionType,
    _ClassInfo,
    format_typehint,
    get_bound_typevars,
    get_type_params,
)
from xorq.vendor.ibis.util import import_object, is_iterable, unalias_package


T_co = TypeVar("T_co", covariant=True)


def as_resolver(obj):
    if callable(obj) and not isinstance(obj, Deferred):
        return Factory(obj)
    else:
        return resolver(obj)


class NoMatch(metaclass=Sentinel):
    """Marker to indicate that a pattern didn't match."""


# TODO(kszucs): have an As[int] or Coerced[int] type in ibis.common.typing which
# would be used to annotate an argument as coercible to int or to a certain type
# without needing for the type to inherit from Coercible
class Pattern(Hashable):
    """Base class for all patterns.

    Patterns are used to match values against a given condition. They are extensively
    used by other core components of Ibis to validate and/or coerce user inputs.
    """

    @classmethod
    def from_typehint(cls, annot: type, allow_coercion: bool = True) -> Pattern:
        """Construct a validator from a python type annotation.

        Parameters
        ----------
        annot
            The typehint annotation to construct the pattern from. This must be
            an already evaluated type annotation.
        allow_coercion
            Whether to use coercion if the typehint is a Coercible type.

        Returns
        -------
        A pattern that matches the given type annotation.

        """
        # TODO(kszucs): cache the result of this function
        # TODO(kszucs): explore issubclass(typ, SupportsInt) etc.
        origin, args = get_origin(annot), get_args(annot)

        if origin is None:
            # the typehint is not generic
            if annot is Ellipsis or annot is AnyType:
                # treat both `Any` and `...` as wildcard
                return _any
            elif isinstance(annot, type):
                # the typehint is a concrete type (e.g. int, str, etc.)
                if allow_coercion and issubclass(annot, Coercible):
                    # the type implements the Coercible protocol so we try to
                    # coerce the value to the given type rather than checking
                    return CoercedTo(annot)
                else:
                    return InstanceOf(annot)
            elif isinstance(annot, TypeVar):
                # if the typehint is a type variable we try to construct a
                # validator from it only if it is covariant and has a bound
                if not annot.__covariant__:
                    raise NotImplementedError(
                        "Only covariant typevars are supported for now"
                    )
                if annot.__bound__:
                    return cls.from_typehint(annot.__bound__)
                else:
                    return _any
            elif isinstance(annot, Enum):
                # for enums we check the value against the enum values
                return EqualTo(annot)
            elif isinstance(annot, str):
                # for strings and forward references we check in a lazy way
                return LazyInstanceOf(annot)
            elif isinstance(annot, ForwardRef):
                return LazyInstanceOf(annot.__forward_arg__)
            else:
                raise TypeError(f"Cannot create validator from annotation {annot!r}")
        elif origin is CoercedTo:
            return CoercedTo(args[0])
        elif origin is Literal:
            # for literal types we check the value against the literal values
            return IsIn(args)
        elif origin is UnionType or origin is Union:
            # this is slightly more complicated because we need to handle
            # Optional[T] which is Union[T, None] and Union[T1, T2, ...]
            *rest, last = args
            if last is type(None):
                # the typehint is Optional[*rest] which is equivalent to
                # Union[*rest, None], so we construct an Option pattern
                if len(rest) == 1:
                    inner = cls.from_typehint(rest[0])
                else:
                    inner = AnyOf(*map(cls.from_typehint, rest))
                return Option(inner)
            else:
                # the typehint is Union[*args] so we construct an AnyOf pattern
                return AnyOf(*map(cls.from_typehint, args))
        elif origin is Annotated:
            # the Annotated typehint can be used to add extra validation logic
            # to the typehint, e.g. Annotated[int, Positive], the first argument
            # is used for isinstance checks, the rest are applied in conjunction
            annot, *extras = args
            return AllOf(cls.from_typehint(annot), *extras)
        elif origin is Callable:
            # the Callable typehint is used to annotate functions, e.g. the
            # following typehint annotates a function that takes two integers
            # and returns a string: Callable[[int, int], str]
            if args:
                # callable with args and return typehints construct a special
                # CallableWith validator
                arg_hints, return_hint = args
                arg_patterns = tuple(map(cls.from_typehint, arg_hints))
                return_pattern = cls.from_typehint(return_hint)
                return CallableWith(arg_patterns, return_pattern)
            else:
                # in case of Callable without args we check for the Callable
                # protocol only
                return InstanceOf(Callable)
        elif issubclass(origin, tuple):
            # construct validators for the tuple elements, but need to treat
            # variadic tuples differently, e.g. tuple[int, ...] is a variadic
            # tuple of integers, while tuple[int] is a tuple with a single int
            first, *rest = args
            if rest == [Ellipsis]:
                return TupleOf(cls.from_typehint(first))
            else:
                return PatternList(map(cls.from_typehint, args), type=origin)
        elif issubclass(origin, Sequence):
            # construct a validator for the sequence elements where all elements
            # must be of the same type, e.g. Sequence[int] is a sequence of ints
            (value_inner,) = map(cls.from_typehint, args)
            if allow_coercion and issubclass(origin, Coercible):
                return GenericSequenceOf(value_inner, type=origin)
            else:
                return SequenceOf(value_inner, type=origin)
        elif issubclass(origin, Mapping):
            # construct a validator for the mapping keys and values, e.g.
            # Mapping[str, int] is a mapping with string keys and int values
            key_inner, value_inner = map(cls.from_typehint, args)
            return MappingOf(key_inner, value_inner, type=origin)
        elif isinstance(origin, GenericMeta):
            # construct a validator for the generic type, see the specific
            # Generic* validators for more details
            if allow_coercion and issubclass(origin, Coercible) and args:
                return GenericCoercedTo(annot)
            else:
                return GenericInstanceOf(annot)
        else:
            raise TypeError(
                f"Cannot create validator from annotation {annot!r} {origin!r}"
            )

    @abstractmethod
    def match(self, value: AnyType, context: dict[str, AnyType]) -> AnyType:
        """Match a value against the pattern.

        Parameters
        ----------
        value
            The value to match the pattern against.
        context
            A dictionary providing arbitrary context for the pattern matching.

        Returns
        -------
        The result of the pattern matching. If the pattern doesn't match
        the value, then it must return the `NoMatch` sentinel value.

        """
        ...

    def describe(self, plural=False):
        return f"matching {self!r}"

    @abstractmethod
    def __eq__(self, other: Pattern) -> bool: ...

    def __invert__(self) -> Not:
        """Syntax sugar for matching the inverse of the pattern."""
        return Not(self)

    def __or__(self, other: Pattern) -> AnyOf:
        """Syntax sugar for matching either of the patterns.

        Parameters
        ----------
        other
            The other pattern to match against.

        Returns
        -------
        New pattern that matches if either of the patterns match.

        """
        return AnyOf(self, other)

    def __and__(self, other: Pattern) -> AllOf:
        """Syntax sugar for matching both of the patterns.

        Parameters
        ----------
        other
            The other pattern to match against.

        Returns
        -------
        New pattern that matches if both of the patterns match.

        """
        return AllOf(self, other)

    def __rshift__(self, other: Deferred) -> Replace:
        """Syntax sugar for replacing a value.

        Parameters
        ----------
        other
            The deferred to use for constructing the replacement value.

        Returns
        -------
        New replace pattern.

        """
        return Replace(self, other)

    def __rmatmul__(self, name: str) -> Capture:
        """Syntax sugar for capturing a value.

        Parameters
        ----------
        name
            The name of the capture.

        Returns
        -------
        New capture pattern.

        """
        return Capture(name, self)

    def __iter__(self) -> SomeOf:
        yield SomeOf(self)


class Is(Slotted, Pattern):
    """Pattern that matches a value against a reference value.

    Parameters
    ----------
    value
        The reference value to match against.

    """

    __slots__ = ("value",)
    value: AnyType

    def match(self, value, context):
        if value is self.value:
            return value
        else:
            return NoMatch


class Any(Slotted, Singleton, Pattern):
    """Pattern that accepts any value, basically a no-op."""

    def match(self, value, context):
        return value


_any = Any()


class Nothing(Slotted, Singleton, Pattern):
    """Pattern that no values."""

    def match(self, value, context):
        return NoMatch


class Capture(Slotted, Pattern):
    """Pattern that captures a value in the context.

    Parameters
    ----------
    pattern
        The pattern to match against.
    key
        The key to use in the context if the pattern matches.

    """

    __slots__ = ("key", "pattern")
    key: AnyType
    pattern: Pattern

    def __init__(self, key, pat=_any):
        if isinstance(key, (Deferred, Resolver)):
            key = as_resolver(key)
            if isinstance(key, Variable):
                key = key.name
            else:
                raise TypeError("Only variables can be used as capture keys")
        super().__init__(key=key, pattern=pattern(pat))

    def match(self, value, context):
        value = self.pattern.match(value, context)
        if value is NoMatch:
            return NoMatch
        context[self.key] = value
        return value


class Replace(Slotted, Pattern):
    """Pattern that replaces a value with the output of another pattern.

    Parameters
    ----------
    matcher
        The pattern to match against.
    replacer
        The deferred to use as a replacement.

    """

    __slots__ = ("matcher", "replacer")
    matcher: Pattern
    replacer: Resolver

    def __init__(self, matcher, replacer):
        super().__init__(matcher=pattern(matcher), replacer=as_resolver(replacer))

    def match(self, value, context):
        value = self.matcher.match(value, context)
        if value is NoMatch:
            return NoMatch
        # use the `_` reserved variable to record the value being replaced
        # in the context, so that it can be used in the replacer pattern
        context["_"] = value
        return self.replacer.resolve(context)


def replace(matcher):
    """More convenient syntax for replacing a value with the output of a function."""

    def decorator(replacer):
        return Replace(matcher, replacer)

    return decorator


class Check(Slotted, Pattern):
    """Pattern that checks a value against a predicate.

    Parameters
    ----------
    predicate
        The predicate to use.

    """

    __slots__ = ("predicate",)
    predicate: Callable

    @classmethod
    def __create__(cls, predicate):
        if isinstance(predicate, (Deferred, Resolver)):
            return DeferredCheck(predicate)
        else:
            return super().__create__(predicate)

    def __init__(self, predicate):
        assert callable(predicate)
        super().__init__(predicate=predicate)

    def describe(self, plural=False):
        if plural:
            return f"values that satisfy {self.predicate.__name__}()"
        else:
            return f"a value that satisfies {self.predicate.__name__}()"

    def match(self, value, context):
        if self.predicate(value):
            return value
        else:
            return NoMatch


class DeferredCheck(Slotted, Pattern):
    __slots__ = ("resolver",)
    resolver: Resolver

    def __init__(self, obj):
        super().__init__(resolver=as_resolver(obj))

    def describe(self, plural=False):
        if plural:
            return f"values that satisfy {self.resolver!r}"
        else:
            return f"a value that satisfies {self.resolver!r}"

    def match(self, value, context):
        context["_"] = value
        if self.resolver.resolve(context):
            return value
        else:
            return NoMatch


class Custom(Slotted, Pattern):
    """User defined custom matcher function.

    Parameters
    ----------
    func
        The function to apply.

    """

    __slots__ = ("func",)
    func: Callable

    def __init__(self, func):
        assert callable(func)
        super().__init__(func=func)

    def match(self, value, context):
        return self.func(value, context)


class EqualTo(Slotted, Pattern):
    """Pattern that checks a value equals to the given value.

    Parameters
    ----------
    value
        The value to check against.

    """

    __slots__ = ("value",)
    value: AnyType

    @classmethod
    def __create__(cls, value):
        if isinstance(value, (Deferred, Resolver)):
            return DeferredEqualTo(value)
        else:
            return super().__create__(value)

    def __init__(self, value):
        super().__init__(value=value)

    def match(self, value, context):
        if value == self.value:
            return value
        else:
            return NoMatch

    def describe(self, plural=False):
        return repr(self.value)


class DeferredEqualTo(Slotted, Pattern):
    """Pattern that checks a value equals to the given value.

    Parameters
    ----------
    value
        The value to check against.

    """

    __slots__ = ("resolver",)
    resolver: Resolver

    def __init__(self, obj):
        super().__init__(resolver=as_resolver(obj))

    def match(self, value, context):
        context["_"] = value
        if value == self.resolver.resolve(context):
            return value
        else:
            return NoMatch

    def describe(self, plural=False):
        return repr(self.resolver)


class Option(Slotted, Pattern):
    """Pattern that matches `None` or a value that passes the inner validator.

    Parameters
    ----------
    pattern
        The inner pattern to use.

    """

    __slots__ = ("pattern", "default")
    pattern: Pattern
    default: AnyType

    def __init__(self, pat, default=None):
        super().__init__(pattern=pattern(pat), default=default)

    def describe(self, plural=False):
        if plural:
            return f"optional {self.pattern.describe(plural=True)}"
        else:
            return f"either None or {self.pattern.describe(plural=False)}"

    def match(self, value, context):
        if value is None:
            if self.default is None:
                return None
            else:
                return self.default
        else:
            return self.pattern.match(value, context)


def _describe_type(typ, plural=False):
    if isinstance(typ, tuple):
        *rest, last = typ
        rest = ", ".join(_describe_type(t, plural=plural) for t in rest)
        last = _describe_type(last, plural=plural)
        return f"{rest} or {last}" if rest else last

    name = format_typehint(typ)
    if plural:
        return f"{name}s"
    elif name[0].lower() in "aeiou":
        return f"an {name}"
    else:
        return f"a {name}"


class TypeOf(Slotted, Pattern):
    """Pattern that matches a value that is of a given type."""

    __slots__ = ("type",)
    type: type

    def __init__(self, typ):
        super().__init__(type=typ)

    def describe(self, plural=False):
        return f"exactly {_describe_type(self.type, plural=plural)}"

    def match(self, value, context):
        if type(value) is self.type:
            return value
        else:
            return NoMatch


class SubclassOf(Slotted, Pattern):
    """Pattern that matches a value that is a subclass of a given type.

    Parameters
    ----------
    type
        The type to check against.

    """

    __slots__ = ("type",)

    def __init__(self, typ):
        super().__init__(type=typ)

    def describe(self, plural=False):
        if plural:
            return f"subclasses of {self.type.__name__}"
        else:
            return f"a subclass of {self.type.__name__}"

    def match(self, value, context):
        if issubclass(value, self.type):
            return value
        else:
            return NoMatch


class InstanceOf(Slotted, Singleton, Pattern):
    """Pattern that matches a value that is an instance of a given type.

    Parameters
    ----------
    types
        The type to check against.

    """

    __slots__ = ("type",)
    type: _ClassInfo

    def __init__(self, typ):
        super().__init__(type=typ)

    def describe(self, plural=False):
        return _describe_type(self.type, plural=plural)

    def match(self, value, context):
        if isinstance(value, self.type):
            return value
        else:
            return NoMatch

    def __call__(self, *args, **kwargs):
        return Object(self.type, *args, **kwargs)


class GenericInstanceOf(Slotted, Pattern):
    """Pattern that matches a value that is an instance of a given generic type.

    Parameters
    ----------
    typ
        The type to check against (must be a generic type).

    Examples
    --------
    >>> class MyNumber(Generic[T_co]):
    ...     value: T_co
    ...
    ...     def __init__(self, value: T_co):
    ...         self.value = value
    ...
    ...     def __eq__(self, other):
    ...         return type(self) is type(other) and self.value == other.value
    >>> p = GenericInstanceOf(MyNumber[int])
    >>> assert p.match(MyNumber(1), {}) == MyNumber(1)
    >>> assert p.match(MyNumber(1.0), {}) is NoMatch
    >>>
    >>> p = GenericInstanceOf(MyNumber[float])
    >>> assert p.match(MyNumber(1.0), {}) == MyNumber(1.0)
    >>> assert p.match(MyNumber(1), {}) is NoMatch

    """

    __slots__ = ("type", "origin", "fields")
    origin: type
    fields: FrozenDict[str, Pattern]

    def __init__(self, typ):
        origin = get_origin(typ)
        typevars = get_bound_typevars(typ)

        fields = {}
        for var, (attr, type_) in typevars.items():
            if not var.__covariant__:
                raise TypeError(
                    f"Typevar {var} is not covariant, cannot use it in a GenericInstanceOf"
                )
            fields[attr] = Pattern.from_typehint(type_, allow_coercion=False)

        super().__init__(type=typ, origin=origin, fields=frozendict(fields))

    def describe(self, plural=False):
        return _describe_type(self.type, plural=plural)

    def match(self, value, context):
        if not isinstance(value, self.origin):
            return NoMatch

        for name, pattern in self.fields.items():
            attr = getattr(value, name)
            if pattern.match(attr, context) is NoMatch:
                return NoMatch

        return value


class LazyInstanceOf(Slotted, Pattern):
    """A version of `InstanceOf` that accepts qualnames instead of imported classes.

    Useful for delaying imports.

    Parameters
    ----------
    types
        The types to check against.

    """

    __fields__ = ("qualname", "package")
    __slots__ = ("qualname", "package", "loaded")
    qualname: str
    package: str
    loaded: type

    def __init__(self, qualname):
        package = unalias_package(qualname.split(".", 1)[0])
        super().__init__(qualname=qualname, package=package)

    def match(self, value, context):
        if hasattr(self, "loaded"):
            return value if isinstance(value, self.loaded) else NoMatch

        for klass in type(value).__mro__:
            package = klass.__module__.split(".", 1)[0]
            if package == self.package:
                typ = import_object(self.qualname)
                object.__setattr__(self, "loaded", typ)
                return value if isinstance(value, typ) else NoMatch

        return NoMatch


class CoercedTo(Slotted, Pattern, Generic[T_co]):
    """Force a value to have a particular Python type.

    If a Coercible subclass is passed, the `__coerce__` method will be used to
    coerce the value. Otherwise, the type will be called with the value as the
    only argument.

    Parameters
    ----------
    type
        The type to coerce to.

    """

    __slots__ = ("type", "func")
    type: T_co

    def __init__(self, type):
        func = type.__coerce__ if issubclass(type, Coercible) else type
        super().__init__(type=type, func=func)

    def describe(self, plural=False):
        type = _describe_type(self.type, plural=False)
        if plural:
            return f"coercibles to {type}"
        else:
            return f"coercible to {type}"

    def match(self, value, context):
        try:
            value = self.func(value)
        except (TypeError, CoercionError):
            return NoMatch

        if isinstance(value, self.type):
            return value
        else:
            return NoMatch

    def __call__(self, *args, **kwargs):
        return Object(self.type, *args, **kwargs)


class GenericCoercedTo(Slotted, Pattern):
    """Force a value to have a particular generic Python type.

    Parameters
    ----------
    typ
        The type to coerce to. Must be a generic type with bound typevars.

    Examples
    --------
    >>> from typing import Generic, TypeVar
    >>>
    >>> T = TypeVar("T", covariant=True)
    >>>
    >>> class MyNumber(Coercible, Generic[T]):
    ...     __slots__ = ("value",)
    ...
    ...     def __init__(self, value):
    ...         self.value = value
    ...
    ...     def __eq__(self, other):
    ...         return type(self) is type(other) and self.value == other.value
    ...
    ...     @classmethod
    ...     def __coerce__(cls, value, T=None):
    ...         if issubclass(T, int):
    ...             return cls(int(value))
    ...         elif issubclass(T, float):
    ...             return cls(float(value))
    ...         else:
    ...             raise CoercionError(f"Cannot coerce to {T}")
    >>> p = GenericCoercedTo(MyNumber[int])
    >>> assert p.match(3.14, {}) == MyNumber(3)
    >>> assert p.match("15", {}) == MyNumber(15)
    >>>
    >>> p = GenericCoercedTo(MyNumber[float])
    >>> assert p.match(3.14, {}) == MyNumber(3.14)
    >>> assert p.match("15", {}) == MyNumber(15.0)

    """

    __slots__ = ("origin", "params", "checker")
    origin: type
    params: FrozenDict[str, type]
    checker: GenericInstanceOf

    def __init__(self, target):
        origin = get_origin(target)
        checker = GenericInstanceOf(target)
        params = frozendict(get_type_params(target))
        super().__init__(origin=origin, params=params, checker=checker)

    def describe(self, plural=False):
        if plural:
            return f"coercibles to {self.checker.describe(plural=False)}"
        else:
            return f"coercible to {self.checker.describe(plural=False)}"

    def match(self, value, context):
        try:
            value = self.origin.__coerce__(value, **self.params)
        except CoercionError:
            return NoMatch

        if self.checker.match(value, context) is NoMatch:
            return NoMatch

        return value


class Not(Slotted, Pattern):
    """Pattern that matches a value that does not match a given pattern.

    Parameters
    ----------
    pattern
        The pattern which the value should not match.

    """

    __slots__ = ("pattern",)
    pattern: Pattern

    def __init__(self, inner):
        super().__init__(pattern=pattern(inner))

    def describe(self, plural=False):
        if plural:
            return f"anything except {self.pattern.describe(plural=True)}"
        else:
            return f"anything except {self.pattern.describe(plural=False)}"

    def match(self, value, context):
        if self.pattern.match(value, context) is NoMatch:
            return value
        else:
            return NoMatch


class AnyOf(Slotted, Pattern):
    """Pattern that if any of the given patterns match.

    Parameters
    ----------
    patterns
        The patterns to match against. The first pattern that matches will be
        returned.

    """

    __slots__ = ("patterns",)
    patterns: tuple[Pattern, ...]

    def __init__(self, *pats):
        patterns = tuple(map(pattern, pats))
        super().__init__(patterns=patterns)

    def describe(self, plural=False):
        *rest, last = self.patterns
        rest = ", ".join(p.describe(plural=plural) for p in rest)
        last = last.describe(plural=plural)
        return f"{rest} or {last}" if rest else last

    def match(self, value, context):
        for pattern in self.patterns:
            result = pattern.match(value, context)
            if result is not NoMatch:
                return result
        return NoMatch


class AllOf(Slotted, Pattern):
    """Pattern that matches if all of the given patterns match.

    Parameters
    ----------
    patterns
        The patterns to match against. The value will be passed through each
        pattern in order. The changes applied to the value propagate through the
        patterns.

    """

    __slots__ = ("patterns",)
    patterns: tuple[Pattern, ...]

    def __init__(self, *pats):
        patterns = tuple(map(pattern, pats))
        super().__init__(patterns=patterns)

    def describe(self, plural=False):
        *rest, last = self.patterns
        rest = ", ".join(p.describe(plural=plural) for p in rest)
        last = last.describe(plural=plural)
        return f"{rest} then {last}" if rest else last

    def match(self, value, context):
        for pattern in self.patterns:
            value = pattern.match(value, context)
            if value is NoMatch:
                return NoMatch
        return value


class Length(Slotted, Pattern):
    """Pattern that matches if the length of a value is within a given range.

    Parameters
    ----------
    exactly
        The exact length of the value. If specified, `at_least` and `at_most`
        must be None.
    at_least
        The minimum length of the value.
    at_most
        The maximum length of the value.

    """

    __slots__ = ("at_least", "at_most")
    at_least: int
    at_most: int

    def __init__(
        self,
        exactly: Optional[int] = None,
        at_least: Optional[int] = None,
        at_most: Optional[int] = None,
    ):
        if exactly is not None:
            if at_least is not None or at_most is not None:
                raise ValueError("Can't specify both exactly and at_least/at_most")
            at_least = exactly
            at_most = exactly
        super().__init__(at_least=at_least, at_most=at_most)

    def describe(self, plural=False):
        if self.at_least is not None and self.at_most is not None:
            if self.at_least == self.at_most:
                return f"with length exactly {self.at_least}"
            else:
                return f"with length between {self.at_least} and {self.at_most}"
        elif self.at_least is not None:
            return f"with length at least {self.at_least}"
        elif self.at_most is not None:
            return f"with length at most {self.at_most}"
        else:
            return "with any length"

    def match(self, value, context):
        length = len(value)
        if self.at_least is not None and length < self.at_least:
            return NoMatch
        if self.at_most is not None and length > self.at_most:
            return NoMatch
        return value


class Between(Slotted, Pattern):
    """Match a value between two bounds.

    Parameters
    ----------
    lower
        The lower bound.
    upper
        The upper bound.

    """

    __slots__ = ("lower", "upper")
    lower: float
    upper: float

    def __init__(self, lower: float = -math.inf, upper: float = math.inf):
        super().__init__(lower=lower, upper=upper)

    def match(self, value, context):
        if self.lower <= value <= self.upper:
            return value
        else:
            return NoMatch


class Contains(Slotted, Pattern):
    """Pattern that matches if a value contains a given value.

    Parameters
    ----------
    needle
        The item that the passed value should contain.

    """

    __slots__ = ("needle",)
    needle: AnyType

    def __init__(self, needle):
        super().__init__(needle=needle)

    def describe(self, plural=False):
        return f"containing {self.needle!r}"

    def match(self, value, context):
        if self.needle in value:
            return value
        else:
            return NoMatch


class IsIn(Slotted, Pattern):
    """Pattern that matches if a value is in a given set.

    Parameters
    ----------
    haystack
        The set of values that the passed value should be in.

    """

    __slots__ = ("haystack",)
    haystack: frozenset

    def __init__(self, haystack):
        super().__init__(haystack=frozenset(haystack))

    def describe(self, plural=False):
        return f"in {set(self.haystack)!r}"

    def match(self, value, context):
        if value in self.haystack:
            return value
        else:
            return NoMatch


class SequenceOf(Slotted, Pattern):
    """Pattern that matches if all of the items in a sequence match a given pattern.

    Specialization of the more flexible GenericSequenceOf pattern which uses two
    additional patterns to possibly coerce the sequence type and to match on
    the length of the sequence.

    Parameters
    ----------
    item
        The pattern to match against each item in the sequence.
    type
        The type to coerce the sequence to. Defaults to tuple.

    """

    __slots__ = ("item", "type")
    item: Pattern
    type: type

    def __init__(self, item, type=list):
        super().__init__(item=pattern(item), type=type)

    def describe(self, plural=False):
        typ = _describe_type(self.type, plural=plural)
        item = self.item.describe(plural=True)
        return f"{typ} of {item}"

    def match(self, values, context):
        if not is_iterable(values):
            return NoMatch

        if self.item == _any:
            # optimization to avoid unnecessary iteration
            result = values
        else:
            result = []
            for item in values:
                item = self.item.match(item, context)
                if item is NoMatch:
                    return NoMatch
                result.append(item)

        return self.type(result)


class GenericSequenceOf(Slotted, Pattern):
    """Pattern that matches if all of the items in a sequence match a given pattern.

    Parameters
    ----------
    item
        The pattern to match against each item in the sequence.
    type
        The type to coerce the sequence to. Defaults to list.
    exactly
        The exact length of the sequence.
    at_least
        The minimum length of the sequence.
    at_most
        The maximum length of the sequence.

    """

    __slots__ = ("item", "type", "length")
    item: Pattern
    type: Pattern
    length: Length

    def __init__(
        self,
        item: Pattern,
        type: type = list,
        exactly: Optional[int] = None,
        at_least: Optional[int] = None,
        at_most: Optional[int] = None,
    ):
        item = pattern(item)
        type = CoercedTo(type)
        length = Length(exactly=exactly, at_least=at_least, at_most=at_most)
        super().__init__(item=item, type=type, length=length)

    def match(self, values, context):
        if not is_iterable(values):
            return NoMatch

        if self.item == _any:
            # optimization to avoid unnecessary iteration
            result = values
        else:
            result = []
            for value in values:
                value = self.item.match(value, context)
                if value is NoMatch:
                    return NoMatch
                result.append(value)

        result = self.type.match(result, context)
        if result is NoMatch:
            return NoMatch

        return self.length.match(result, context)


class GenericMappingOf(Slotted, Pattern):
    """Pattern that matches if all of the keys and values match the given patterns.

    Parameters
    ----------
    key
        The pattern to match the keys against.
    value
        The pattern to match the values against.
    type
        The type to coerce the mapping to. Defaults to dict.

    """

    __slots__ = ("key", "value", "type")
    key: Pattern
    value: Pattern
    type: Pattern

    def __init__(self, key: Pattern, value: Pattern, type: type = dict):
        super().__init__(key=pattern(key), value=pattern(value), type=CoercedTo(type))

    def match(self, value, context):
        if not isinstance(value, Mapping):
            return NoMatch

        result = {}
        for k, v in value.items():
            if (k := self.key.match(k, context)) is NoMatch:
                return NoMatch
            if (v := self.value.match(v, context)) is NoMatch:
                return NoMatch
            result[k] = v

        result = self.type.match(result, context)
        if result is NoMatch:
            return NoMatch

        return result


MappingOf = GenericMappingOf


class Attrs(Slotted, Pattern):
    __slots__ = ("fields",)
    fields: FrozenDict[str, Pattern]

    def __init__(self, **fields):
        fields = frozendict(toolz.valmap(pattern, fields))
        super().__init__(fields=fields)

    def match(self, value, context):
        for attr, pattern in self.fields.items():
            if not hasattr(value, attr):
                return NoMatch

            v = getattr(value, attr)
            if match(pattern, v, context) is NoMatch:
                return NoMatch

        return value


class Object(Slotted, Pattern):
    """Pattern that matches if the object has the given attributes and they match the given patterns.

    The type must conform the structural pattern matching protocol, e.g. it must have a
    __match_args__ attribute that is a tuple of the names of the attributes to match.

    Parameters
    ----------
    type
        The type of the object.
    *args
        The positional arguments to match against the attributes of the object.
    **kwargs
        The keyword arguments to match against the attributes of the object.

    """

    __slots__ = ("type", "args", "kwargs")
    type: Pattern
    args: tuple[Pattern, ...]
    kwargs: FrozenDict[str, Pattern]

    @classmethod
    def __create__(cls, type, *args, **kwargs):
        if not args and not kwargs:
            return InstanceOf(type)
        return super().__create__(type, *args, **kwargs)

    def __init__(self, typ, *args, **kwargs):
        if isinstance(typ, type) and len(typ.__match_args__) < len(args):
            raise ValueError(
                "The type to match has fewer `__match_args__` than the number "
                "of positional arguments in the pattern"
            )
        typ = pattern(typ)
        args = tuple(map(pattern, args))
        kwargs = frozendict(toolz.valmap(pattern, kwargs))
        super().__init__(type=typ, args=args, kwargs=kwargs)

    def match(self, value, context):
        if self.type.match(value, context) is NoMatch:
            return NoMatch

        # the pattern requirest more positional arguments than the object has
        if len(value.__match_args__) < len(self.args):
            return NoMatch
        patterns = dict(zip(value.__match_args__, self.args))
        patterns.update(self.kwargs)

        fields = {}
        changed = False
        for name, pattern in patterns.items():
            try:
                attr = getattr(value, name)
            except AttributeError:
                return NoMatch

            result = pattern.match(attr, context)
            if result is NoMatch:
                return NoMatch
            elif result != attr:
                changed = True
                fields[name] = result
            else:
                fields[name] = attr

        if changed:
            return type(value)(**fields)
        else:
            return value


class Node(Slotted, Pattern):
    __slots__ = ("type", "each_arg")
    type: Pattern

    def __init__(self, type, each_arg):
        super().__init__(type=pattern(type), each_arg=pattern(each_arg))

    def match(self, value, context):
        if self.type.match(value, context) is NoMatch:
            return NoMatch

        newargs = {}
        changed = False
        for name, arg in zip(value.__argnames__, value.__args__):
            result = self.each_arg.match(arg, context)
            if result is NoMatch:
                newargs[name] = arg
            else:
                newargs[name] = result
                changed = True

        if changed:
            return value.__class__(**newargs)
        else:
            return value


class CallableWith(Slotted, Pattern):
    __slots__ = ("args", "return_")
    args: tuple
    return_: AnyType

    def __init__(self, args, return_=_any):
        super().__init__(args=tuple(args), return_=return_)

    def match(self, value, context):
        from xorq.vendor.ibis.common.annotations import EMPTY, annotated

        if not callable(value):
            return NoMatch

        fn = annotated(self.args, self.return_, value)

        has_varargs = False
        positional, required_positional = [], []
        for p in fn.__signature__.parameters.values():
            if p.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD):
                positional.append(p)
                if p.default is EMPTY:
                    required_positional.append(p)
            elif p.kind is Parameter.KEYWORD_ONLY and p.default is EMPTY:
                raise TypeError(
                    "Callable has mandatory keyword-only arguments which cannot be specified"
                )
            elif p.kind is Parameter.VAR_POSITIONAL:
                has_varargs = True

        if len(required_positional) > len(self.args):
            # Callable has more positional arguments than expected")
            return NoMatch
        elif len(positional) < len(self.args) and not has_varargs:
            # Callable has less positional arguments than expected")
            return NoMatch
        else:
            return fn


class SomeOf(Slotted, Pattern):
    __slots__ = ("pattern", "delimiter")

    @classmethod
    def __create__(cls, *args, **kwargs):
        if len(args) == 1:
            return super().__create__(*args, **kwargs)
        else:
            return SomeChunksOf(*args, **kwargs)

    def __init__(self, item, **kwargs):
        pattern = GenericSequenceOf(item, **kwargs)
        delimiter = pattern.item
        super().__init__(pattern=pattern, delimiter=delimiter)

    def match(self, values, context):
        return self.pattern.match(values, context)


class SomeChunksOf(Slotted, Pattern):
    """Pattern that unpacks a value into its elements.

    Designed to be used inside a `PatternList` pattern with the `*` syntax.
    """

    __slots__ = ("pattern", "delimiter")

    def __init__(self, *args, **kwargs):
        pattern = GenericSequenceOf(PatternList(args), **kwargs)
        delimiter = pattern.item.patterns[0]
        super().__init__(pattern=pattern, delimiter=delimiter)

    def chunk(self, values, context):
        chunk = []
        for item in values:
            if self.delimiter.match(item, context) is NoMatch:
                chunk.append(item)
            else:
                if chunk:  # only yield if there are items in the chunk
                    yield chunk
                chunk = [item]  # start a new chunk with the delimiter
        if chunk:
            yield chunk

    def match(self, values, context):
        chunks = self.chunk(values, context)
        result = self.pattern.match(chunks, context)
        if result is NoMatch:
            return NoMatch
        else:
            return [el for lst in result for el in lst]


def _maybe_unwrap_capture(obj):
    return obj.pattern if isinstance(obj, Capture) else obj


class PatternList(Slotted, Pattern):
    """Pattern that matches if the respective items in a tuple match the given patterns.

    Parameters
    ----------
    fields
        The patterns to match the respective items in the tuple.

    """

    __slots__ = ("patterns", "type")
    patterns: tuple[Pattern, ...]
    type: type

    @classmethod
    def __create__(cls, patterns, type=list):
        if patterns == ():
            return EqualTo(patterns)

        patterns = tuple(map(pattern, patterns))
        for pat in patterns:
            pat = _maybe_unwrap_capture(pat)
            if isinstance(pat, (SomeOf, SomeChunksOf)):
                return VariadicPatternList(patterns, type)

        return super().__create__(patterns, type)

    def __init__(self, patterns, type):
        super().__init__(patterns=patterns, type=type)

    def describe(self, plural=False):
        patterns = ", ".join(f.describe(plural=False) for f in self.patterns)
        if plural:
            return f"tuples of ({patterns})"
        else:
            return f"a tuple of ({patterns})"

    def match(self, values, context):
        if not is_iterable(values):
            return NoMatch

        if len(values) != len(self.patterns):
            return NoMatch

        result = []
        for pattern, value in zip(self.patterns, values):
            value = pattern.match(value, context)
            if value is NoMatch:
                return NoMatch
            result.append(value)

        return self.type(result)


class VariadicPatternList(Slotted, Pattern):
    __slots__ = ("patterns", "type")
    patterns: tuple[Pattern, ...]
    type: type

    def __init__(self, patterns, type=list):
        patterns = tuple(map(pattern, patterns))
        super().__init__(patterns=patterns, type=type)

    def match(self, value, context):
        if not self.patterns:
            return NoMatch if value else []

        it = RewindableIterator(value)
        result = []

        following_patterns = self.patterns[1:] + (Nothing(),)
        for current, following in zip(self.patterns, following_patterns):
            original = current
            current = _maybe_unwrap_capture(current)
            following = _maybe_unwrap_capture(following)

            if isinstance(current, (SomeOf, SomeChunksOf)):
                if isinstance(following, (SomeOf, SomeChunksOf)):
                    following = following.delimiter

                matches = []
                while True:
                    it.checkpoint()
                    try:
                        item = next(it)
                    except StopIteration:
                        break

                    res = following.match(item, context)
                    if res is NoMatch:
                        matches.append(item)
                    else:
                        it.rewind()
                        break

                res = original.match(matches, context)
                if res is NoMatch:
                    return NoMatch
                else:
                    result.extend(res)
            else:
                try:
                    item = next(it)
                except StopIteration:
                    return NoMatch

                res = original.match(item, context)
                if res is NoMatch:
                    return NoMatch
                else:
                    result.append(res)

        return self.type(result)


def NoneOf(*args) -> Pattern:
    """Match none of the passed patterns."""
    return Not(AnyOf(*args))


def ListOf(pattern):
    """Match a list of items matching the given pattern."""
    return SequenceOf(pattern, type=list)


def TupleOf(pattern):
    """Match a variable-length tuple of items matching the given pattern."""
    return SequenceOf(pattern, type=tuple)


def DictOf(key_pattern, value_pattern):
    """Match a dictionary with keys and values matching the given patterns."""
    return MappingOf(key_pattern, value_pattern, type=dict)


def FrozenDictOf(key_pattern, value_pattern):
    """Match a frozendict with keys and values matching the given patterns."""
    return MappingOf(key_pattern, value_pattern, type=frozendict)


def pattern(obj: AnyType) -> Pattern:
    """Create a pattern from various types.

    Not that if a Coercible type is passed as argument, the constructed pattern
    won't attempt to coerce the value during matching. In order to allow type
    coercions use `Pattern.from_typehint()` factory method.

    Parameters
    ----------
    obj
        The object to create a pattern from. Can be a pattern, a type, a callable,
        a mapping, an iterable or a value.

    Examples
    --------
    >>> assert pattern(Any()) == Any()
    >>> assert pattern(int) == InstanceOf(int)
    >>>
    >>> @pattern
    ... def as_int(x, context):
    ...     return int(x)
    >>>
    >>> assert as_int.match(1, {}) == 1

    Returns
    -------
    The constructed pattern.

    """
    if obj is Ellipsis:
        return _any
    elif isinstance(obj, Pattern):
        return obj
    elif isinstance(obj, (Deferred, Resolver)):
        return Capture(obj)
    elif isinstance(obj, Mapping):
        return EqualTo(FrozenDict(obj))
    elif isinstance(obj, Sequence):
        if isinstance(obj, (str, bytes)):
            return EqualTo(obj)
        else:
            return PatternList(obj, type=type(obj))
    elif isinstance(obj, type):
        return InstanceOf(obj)
    elif get_origin(obj):
        return Pattern.from_typehint(obj, allow_coercion=False)
    elif callable(obj):
        return Custom(obj)
    else:
        return EqualTo(obj)


def match(
    pat: Pattern, value: AnyType, context: Optional[dict[str, AnyType]] = None
) -> Any:
    """Match a value against a pattern.

    Parameters
    ----------
    pat
        The pattern to match against.
    value
        The value to match.
    context
        Arbitrary mapping of values to be used while matching.

    Returns
    -------
    The matched value if the pattern matches, otherwise :obj:`NoMatch`.

    Examples
    --------
    >>> assert match(Any(), 1) == 1
    >>> assert match(1, 1) == 1
    >>> assert match(1, 2) is NoMatch
    >>> assert match(1, 1, context={"x": 1}) == 1
    >>> assert match(1, 2, context={"x": 1}) is NoMatch
    >>> assert match([1, int], [1, 2]) == [1, 2]
    >>> assert match([1, int, "a" @ InstanceOf(str)], [1, 2, "three"]) == [
    ...     1,
    ...     2,
    ...     "three",
    ... ]

    """
    if context is None:
        context = {}

    pat = pattern(pat)
    result = pat.match(value, context)
    return NoMatch if result is NoMatch else result


IsTruish = Check(lambda x: bool(x))
IsNumber = InstanceOf(numbers.Number) & ~InstanceOf(bool)
IsString = InstanceOf(str)

As = CoercedTo
Eq = EqualTo
In = IsIn
If = Check
Some = SomeOf



"""Lower the ibis expression graph to a SQL-like relational algebra."""

from __future__ import annotations

import operator
from collections.abc import Mapping
from functools import reduce
from typing import TYPE_CHECKING, Any

import toolz
from public import public

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.common.collections import FrozenDict  # noqa: TCH001
from xorq.vendor.ibis.common.deferred import var
from xorq.vendor.ibis.common.graph import Graph
from xorq.vendor.ibis.common.patterns import InstanceOf, Object, Pattern, replace
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001
from xorq.vendor.ibis.expr.rewrites import d, p, replace_parameter
from xorq.vendor.ibis.expr.schema import Schema


if TYPE_CHECKING:
    from collections.abc import Sequence

x = var("x")
y = var("y")


@public
class CTE(ops.Relation):
    """Common table expression."""

    parent: ops.Relation

    @attribute
    def schema(self):
        return self.parent.schema

    @attribute
    def values(self):
        return self.parent.values


@public
class Select(ops.Relation):
    """Relation modelled after SQL's SELECT statement."""

    parent: ops.Relation
    selections: FrozenDict[str, ops.Value] = {}
    predicates: VarTuple[ops.Value[dt.Boolean]] = ()
    qualified: VarTuple[ops.Value[dt.Boolean]] = ()
    sort_keys: VarTuple[ops.SortKey] = ()
    distinct: bool = False

    def is_star_selection(self):
        return tuple(self.values.items()) == tuple(self.parent.fields.items())

    @attribute
    def values(self):
        return self.selections

    @attribute
    def schema(self):
        return Schema({k: v.dtype for k, v in self.selections.items()})


@public
class FirstValue(ops.Analytic):
    """Retrieve the first element."""

    arg: ops.Column[dt.Any]

    @attribute
    def dtype(self):
        return self.arg.dtype


@public
class LastValue(ops.Analytic):
    """Retrieve the last element."""

    arg: ops.Column[dt.Any]

    @attribute
    def dtype(self):
        return self.arg.dtype


# TODO(kszucs): there is a better strategy to rewrite the relational operations
# to Select nodes by wrapping the leaf nodes in a Select node and then merging
# Project, Filter, Sort, etc. incrementally into the Select node. This way we
# can have tighter control over simplification logic.


@replace(p.Project)
def project_to_select(_, **kwargs):
    """Convert a Project node to a Select node."""
    return Select(_.parent, selections=_.values)


def partition_predicates(predicates):
    qualified = []
    unqualified = []

    for predicate in predicates:
        if predicate.find(ops.WindowFunction, filter=ops.Value):
            qualified.append(predicate)
        else:
            unqualified.append(predicate)

    return unqualified, qualified


@replace(p.Filter)
def filter_to_select(_, **kwargs):
    """Convert a Filter node to a Select node."""
    predicates, qualified = partition_predicates(_.predicates)
    return Select(
        _.parent, selections=_.values, predicates=predicates, qualified=qualified
    )


@replace(p.Sort)
def sort_to_select(_, **kwargs):
    """Convert a Sort node to a Select node."""
    return Select(_.parent, selections=_.values, sort_keys=_.keys)


@replace(p.Distinct)
def distinct_to_select(_, **kwargs):
    """Convert a Distinct node to a Select node."""
    return Select(_.parent, selections=_.values, distinct=True)


@replace(p.DropColumns)
def drop_columns_to_select(_, **kwargs):
    """Convert a DropColumns node to a Select node."""
    # if we're dropping fewer than 50% of the parent table's columns then the
    # compiled query will likely be smaller than if we list everything *NOT*
    # being dropped
    if len(_.columns_to_drop) < len(_.schema) // 2:
        return _
    return Select(_.parent, selections=_.values)


@replace(p.FillNull)
def fill_null_to_select(_, **kwargs):
    """Rewrite FillNull to a Select node."""
    if isinstance(_.replacements, Mapping):
        mapping = _.replacements
    else:
        mapping = {
            name: _.replacements
            for name, type in _.parent.schema.items()
            if type.nullable
        }

    if not mapping:
        return _.parent

    selections = {}
    for name in _.parent.schema.names:
        col = ops.Field(_.parent, name)
        if (value := mapping.get(name)) is not None:
            col = ops.Coalesce((col, value))
        selections[name] = col

    return Select(_.parent, selections=selections)


@replace(p.DropNull)
def drop_null_to_select(_, **kwargs):
    """Rewrite DropNull to a Select node."""
    if _.subset is None:
        columns = [ops.Field(_.parent, name) for name in _.parent.schema.names]
    else:
        columns = _.subset

    if columns:
        preds = [
            reduce(
                ops.And if _.how == "any" else ops.Or,
                [ops.NotNull(c) for c in columns],
            )
        ]
    elif _.how == "all":
        preds = [ops.Literal(False, dtype=dt.bool)]
    else:
        return _.parent

    return Select(_.parent, selections=_.values, predicates=tuple(preds))


@replace(p.WindowFunction(p.First | p.Last))
def first_to_firstvalue(_, **kwargs):
    """Convert a First or Last node to a FirstValue or LastValue node."""
    if _.func.where is not None:
        raise com.UnsupportedOperationError(
            f"`{type(_.func).__name__.lower()}` with `where` is unsupported "
            "in a window function"
        )
    klass = FirstValue if isinstance(_.func, ops.First) else LastValue
    return _.copy(func=klass(_.func.arg))


@replace(p.Alias)
def remove_aliases(_, **kwargs):
    """Remove all remaining aliases, they're not needed for remaining compilation."""
    return _.arg


def complexity(node):
    """Assign a complexity score to a node.

    Subsequent projections can be merged into a single projection by replacing
    the fields referenced in the outer projection with the computed expressions
    from the inner projection. This inlining can result in very complex value
    expressions depending on the projections. In order to prevent excessive
    inlining, we assign a complexity score to each node.

    The complexity score assigns 1 to each value expression and adds up in the
    tree hierarchy unless there is a Field node where we don't add up the
    complexity of the referenced relation. This way we treat fields kind of like
    reusable variables considering them less complex than they were inlined.
    """

    def accum(node, *args):
        if isinstance(node, ops.Field):
            return 1
        else:
            return 1 + sum(args)

    return node.map_nodes(accum)[node]


@replace(Object(Select, Object(Select)))
def merge_select_select(_, **kwargs):
    """Merge subsequent Select relations into one.

    This rewrites eliminates `_.parent` by merging the outer and the inner
    `predicates`, `sort_keys` and keeping the outer `selections`. All selections
    from the inner Select are inlined into the outer Select.
    """
    # don't merge if either the outer or the inner select has window functions
    blocking = (
        ops.WindowFunction,
        ops.ExistsSubquery,
        ops.InSubquery,
        ops.Unnest,
        ops.Impure,
    )
    if _.find_below(blocking, filter=ops.Value):
        return _
    if _.parent.find_below(blocking, filter=ops.Value):
        return _

    if _.parent.distinct:
        # The inner query is distinct.
        #
        # If the outer query is distinct, it's only safe to merge if it's a simple subselection:
        # - Fusing in the presence of non-deterministic calls in the select would lead to
        #   incorrect results
        # - Fusing in the presence of expensive calls in the select would lead to potential
        #   performance pitfalls
        if _.distinct and not all(
            isinstance(v, ops.Field) for v in _.selections.values()
        ):
            return _

        # If the outer query isn't distinct, it's only safe to merge if the outer is a SELECT *:
        # - If new columns are added, they might be non-distinct, changing the distinctness
        # - If previous columns are removed, that would also change the distinctness
        if not _.distinct and not _.is_star_selection():
            return _

        distinct = True
    elif _.distinct:
        # The outer query is distinct and the inner isn't. It's only safe to merge if either
        # - The inner query isn't ordered
        # - The outer query is a SELECT *
        #
        # Otherwise we run the risk that the outer query drops columns needed for the ordering of
        # the inner query - many backends don't allow select distinc queries to order by columns
        # that aren't present in their selection, like
        #
        #   SELECT DISTINCT a, b FROM t ORDER BY c  --- some backends will explode at this
        #
        # An alternate solution would be to drop the inner ORDER BY clause, since the backend will
        # ignore it anyway since it's a subquery. That feels potentially risky though, better
        # to generate the SQL as written.
        if _.parent.sort_keys and not _.is_star_selection():
            return _

        distinct = True
    else:
        # Neither query is distinct, safe to merge
        distinct = False

    subs = {ops.Field(_.parent, k): v for k, v in _.parent.values.items()}
    selections = {k: v.replace(subs, filter=ops.Value) for k, v in _.selections.items()}

    predicates = tuple(p.replace(subs, filter=ops.Value) for p in _.predicates)
    unique_predicates = toolz.unique(_.parent.predicates + predicates)

    qualified = tuple(p.replace(subs, filter=ops.Value) for p in _.qualified)
    unique_qualified = toolz.unique(_.parent.qualified + qualified)

    sort_keys = tuple(s.replace(subs, filter=ops.Value) for s in _.sort_keys)
    sort_key_exprs = {s.expr for s in sort_keys}
    parent_sort_keys = tuple(
        k for k in _.parent.sort_keys if k.expr not in sort_key_exprs
    )
    unique_sort_keys = sort_keys + parent_sort_keys

    result = Select(
        _.parent.parent,
        selections=selections,
        predicates=unique_predicates,
        qualified=unique_qualified,
        sort_keys=unique_sort_keys,
        distinct=distinct,
    )
    return result if complexity(result) <= complexity(_) else _


def extract_ctes(node: ops.Relation) -> set[ops.Relation]:
    cte_types = (Select, ops.Aggregate, ops.JoinChain, ops.Set, ops.Limit, ops.Sample)
    dont_count = (ops.Field, ops.CountStar, ops.CountDistinctStar)

    g = Graph.from_bfs(node, filter=~InstanceOf(dont_count))
    result = set()
    for op, dependents in g.invert().items():
        if isinstance(op, ops.View) or (
            len(dependents) > 1 and isinstance(op, cte_types)
        ):
            result.add(op)

    return result


def sqlize(
    node: ops.Node,
    params: Mapping[ops.ScalarParameter, Any],
    rewrites: Sequence[Pattern] = (),
    post_rewrites: Sequence[Pattern] = (),
    fuse_selects: bool = True,
) -> tuple[ops.Node, list[ops.Node]]:
    """Lower the ibis expression graph to a SQL-like relational algebra.

    Parameters
    ----------
    node
        The root node of the expression graph.
    params
        A mapping of scalar parameters to their values.
    rewrites
        Supplementary rewrites to apply before SQL-specific transforms.
    post_rewrites
        Supplementary rewrites to apply after SQL-specific transforms.
    fuse_selects
        Whether to merge subsequent Select nodes into one where possible.

    Returns
    -------
    Tuple of the rewritten expression graph and a list of CTEs.

    """
    assert isinstance(node, ops.Relation)

    # apply the backend specific rewrites
    if rewrites:
        node = node.replace(reduce(operator.or_, rewrites))

    # lower the expression graph to a SQL-like relational algebra
    context = {"params": params}
    result = node.replace(
        replace_parameter
        | remove_aliases
        | project_to_select
        | filter_to_select
        | sort_to_select
        | distinct_to_select
        | fill_null_to_select
        | drop_null_to_select
        | drop_columns_to_select
        | first_to_firstvalue,
        context=context,
    )

    # squash subsequent Select nodes into one
    if fuse_selects:
        result = result.replace(merge_select_select)

    if post_rewrites:
        result = result.replace(reduce(operator.or_, post_rewrites))

    # extract common table expressions while wrapping them in a CTE node
    ctes = extract_ctes(result)

    if ctes:

        def apply_ctes(node, kwargs):
            new = node.__recreate__(kwargs) if kwargs else node
            return CTE(new) if node in ctes else new

        result = result.replace(apply_ctes)
        return result, [cte.parent for cte in result.find(CTE, ordered=True)]
    return result, []


# supplemental rewrites selectively used on a per-backend basis


@replace(Select)
def split_select_distinct_with_order_by(_):
    """Split a `SELECT DISTINCT ... ORDER BY` query when needed.

    Some databases (postgres, pyspark, ...) have issues with two types of
    ordered select distinct statements:

    ```
    --- ORDER BY with an expression instead of a name in the select list
    SELECT DISTINCT a, b FROM t ORDER BY a + 1

    --- ORDER BY using a qualified column name, rather than the alias in the select list
    SELECT DISTINCT a, b as x FROM t ORDER BY b  --- or t.b
    ```

    We solve both these cases by splitting everything except the `ORDER BY`
    into a subquery.

    ```
    SELECT DISTINCT a, b FROM t WHERE a > 10 ORDER BY a + 1
    --- is rewritten as ->
    SELECT * FROM (SELECT DISTINCT a, b FROM t WHERE a > 10) ORDER BY a + 1
    ```
    """
    # risingwave and pyspark also don't allow qualified names as sort keys, like
    #   SELECT DISTINCT t.a FROM t ORDER BY t.a
    # To avoid having specific rewrite rules for these backends to use only
    # local names, we always split SELECT DISTINCT from ORDER BY here. Otherwise we
    # could also avoid splitting if all sort keys appear in the select list.
    if _.distinct and _.sort_keys:
        inner = _.copy(sort_keys=())
        subs = {v: ops.Field(inner, k) for k, v in inner.values.items()}
        sort_keys = tuple(s.replace(subs, filter=ops.Value) for s in _.sort_keys)
        selections = {
            k: v.replace(subs, filter=ops.Value) for k, v in _.selections.items()
        }
        return Select(inner, selections=selections, sort_keys=sort_keys)
    return _


@replace(p.WindowFunction(func=p.NTile(y), order_by=()))
def add_order_by_to_empty_ranking_window_functions(_, **kwargs):
    """Add an ORDER BY clause to rank window functions that don't have one."""
    return _.copy(order_by=(y,))


"""Replace checks against an empty right side with `False`."""
empty_in_values_right_side = p.InValues(options=()) >> d.Literal(False, dtype=dt.bool)


@replace(
    p.WindowFunction(p.RankBase | p.NTile)
    | p.StringFind
    | p.FindInSet
    | p.ArrayPosition
)
def one_to_zero_index(_, **kwargs):
    """Subtract one from one-index functions."""
    return ops.Subtract(_, 1)


@replace(ops.NthValue)
def add_one_to_nth_value_input(_, **kwargs):
    if isinstance(_.nth, ops.Literal):
        nth = ops.Literal(_.nth.value + 1, dtype=_.nth.dtype)
    else:
        nth = ops.Add(_.nth, 1)
    return _.copy(nth=nth)


@replace(p.WindowFunction(order_by=()))
def rewrite_empty_order_by_window(_, **kwargs):
    return _.copy(order_by=(ops.NULL,))


@replace(p.WindowFunction(p.RowNumber | p.NTile))
def exclude_unsupported_window_frame_from_row_number(_, **kwargs):
    return ops.Subtract(_.copy(start=None, end=0), 1)


@replace(p.WindowFunction(p.MinRank | p.DenseRank, start=None))
def exclude_unsupported_window_frame_from_rank(_, **kwargs):
    return ops.Subtract(
        _.copy(start=None, end=0, order_by=_.order_by or (ops.NULL,)), 1
    )


@replace(
    p.WindowFunction(
        p.Lag | p.Lead | p.PercentRank | p.CumeDist | p.Any | p.All, start=None
    )
)
def exclude_unsupported_window_frame_from_ops(_, **kwargs):
    return _.copy(start=None, end=0, order_by=_.order_by or (ops.NULL,))


# Rewrite rules for lowering a high-level operation into one composed of more
# primitive operations.


@replace(p.Log2)
def lower_log2(_, **kwargs):
    """Rewrite `log2` as `log`."""
    return ops.Log(_.arg, base=2)


@replace(p.Log10)
def lower_log10(_, **kwargs):
    """Rewrite `log10` as `log`."""
    return ops.Log(_.arg, base=10)


@replace(p.Bucket)
def lower_bucket(_, **kwargs):
    """Rewrite `Bucket` as `SearchedCase`."""
    cases = []
    results = []

    if _.closed == "left":
        l_cmp = ops.LessEqual
        r_cmp = ops.Less
    else:
        l_cmp = ops.Less
        r_cmp = ops.LessEqual

    user_num_buckets = len(_.buckets) - 1

    bucket_id = 0
    if _.include_under:
        if user_num_buckets > 0:
            cmp = ops.Less if _.close_extreme else r_cmp
        else:
            cmp = ops.LessEqual if _.closed == "right" else ops.Less
        cases.append(cmp(_.arg, _.buckets[0]))
        results.append(bucket_id)
        bucket_id += 1

    for j, (lower, upper) in enumerate(zip(_.buckets, _.buckets[1:])):
        if _.close_extreme and (
            (_.closed == "right" and j == 0)
            or (_.closed == "left" and j == (user_num_buckets - 1))
        ):
            cases.append(
                ops.And(ops.LessEqual(lower, _.arg), ops.LessEqual(_.arg, upper))
            )
            results.append(bucket_id)
        else:
            cases.append(ops.And(l_cmp(lower, _.arg), r_cmp(_.arg, upper)))
            results.append(bucket_id)
        bucket_id += 1

    if _.include_over:
        if user_num_buckets > 0:
            cmp = ops.Less if _.close_extreme else l_cmp
        else:
            cmp = ops.Less if _.closed == "right" else ops.LessEqual

        cases.append(cmp(_.buckets[-1], _.arg))
        results.append(bucket_id)
        bucket_id += 1

    return ops.SearchedCase(
        cases=tuple(cases), results=tuple(results), default=ops.NULL
    )


@replace(p.Capitalize)
def lower_capitalize(_, **kwargs):
    """Rewrite Capitalize in terms of substring, concat, upper, and lower."""
    first = ops.Uppercase(ops.Substring(_.arg, start=0, length=1))
    # use length instead of length - 1 to avoid backends complaining about
    # asking for negative length
    #
    # there are at most length - 1 characters, so asking for length is fine
    rest = ops.Lowercase(ops.Substring(_.arg, start=1, length=ops.StringLength(_.arg)))
    return ops.StringConcat((first, rest))


@replace(p.Sample)
def lower_sample(_, **kwargs):
    """Rewrite Sample as `t.filter(random() <= fraction)`.

    Errors as unsupported if a `seed` is specified.
    """
    if _.seed is not None:
        raise com.UnsupportedOperationError(
            "`Table.sample` with a random seed is unsupported"
        )
    return ops.Filter(_.parent, (ops.LessEqual(ops.RandomScalar(), _.fraction),))



import pathlib

import pytest
import structlog
from structlog.testing import LogCapture

from xorq.common.utils.logging_utils import (
    get_log_path,
    log_initial_state,
)


@pytest.fixture(name="log_output")
def fixture_log_output():
    return LogCapture()


@pytest.fixture(autouse=True)
def fixture_configure_structlog(log_output):
    structlog.configure(processors=[log_output])


def _has_event(events, event_name):
    return any(entry.get("event", "") == event_name for entry in events)


def test_logging_with_git(log_output):
    log_initial_state()

    assert _has_event(log_output.entries, "git state")
    assert not _has_event(log_output.entries, "xorq version")


def test_logging_without_git(log_output, tmp_path):
    log_initial_state(cwd=tmp_path)

    assert not _has_event(log_output.entries, "git state")
    assert _has_event(log_output.entries, "xorq version")


def test_temp_log_path():
    bad_log_path = "/nonexistantandunwritablepath/"
    log_path = pathlib.Path(get_log_path(bad_log_path))
    assert log_path.exists()
    with pytest.raises(ValueError):
        log_path.relative_to(bad_log_path)



from __future__ import annotations

import contextlib
import math
from copy import deepcopy

import sqlglot.expressions as sge
import sqlglot.generator as sgn
from sqlglot import transforms
from sqlglot.dialects import (
    TSQL,
    Hive,
    MySQL,
    Oracle,
    Postgres,
    Snowflake,
    Spark,
    SQLite,
    Trino,
)
from sqlglot.dialects import ClickHouse as _ClickHouse
from sqlglot.dialects.dialect import rename_func
from sqlglot.helper import find_new_name, seq_get


class ClickHouse(_ClickHouse):
    class Generator(_ClickHouse.Generator):
        _ClickHouse.Generator.TRANSFORMS |= {
            sge.ArraySize: rename_func("length"),
            sge.ArraySort: rename_func("arraySort"),
            sge.LogicalAnd: rename_func("min"),
            sge.LogicalOr: rename_func("max"),
        }

        def except_op(self, expression: sge.Except) -> str:
            return f"EXCEPT{' DISTINCT' if expression.args.get('distinct') else ' ALL'}"

        def intersect_op(self, expression: sge.Intersect) -> str:
            return (
                f"INTERSECT{' DISTINCT' if expression.args.get('distinct') else ' ALL'}"
            )


with contextlib.suppress(AttributeError):
    ClickHouse.Generator.TRANSFORMS[sge.Median] = rename_func("median")


class DataFusion(Postgres):
    class Generator(Postgres.Generator):
        TRANSFORMS = Postgres.Generator.TRANSFORMS.copy() | {
            sge.Select: transforms.preprocess([transforms.eliminate_qualify]),
            sge.Pow: rename_func("pow"),
            sge.IsNan: rename_func("isnan"),
            sge.CurrentTimestamp: rename_func("now"),
            sge.CurrentDate: rename_func("today"),
            sge.Split: rename_func("string_to_array"),
            sge.Array: rename_func("make_array"),
            sge.ArrayContains: rename_func("array_has"),
            sge.ArraySize: rename_func("array_length"),
        }


with contextlib.suppress(AttributeError):
    DataFusion.Generator.TRANSFORMS[sge.Median] = rename_func("median")


class Druid(Postgres):
    class Generator(Postgres.Generator):
        TRANSFORMS = Postgres.Generator.TRANSFORMS.copy() | {
            sge.ApproxDistinct: rename_func("approx_count_distinct"),
            sge.Pow: rename_func("power"),
        }


def _interval(self, e, quote_arg=True):
    """Work around the inability to handle string literals in INTERVAL syntax."""
    arg = e.args["this"].this
    with contextlib.suppress(AttributeError):
        arg = arg.sql(self.dialect)

    if quote_arg:
        arg = f"'{arg}'"

    return f"INTERVAL {arg} {e.args['unit']}"


def _group_concat(self, e):
    this = self.sql(e, "this")
    separator = self.sql(e, "separator") or "','"
    return f"GROUP_CONCAT({this} SEPARATOR {separator})"


class Exasol(Postgres):
    class Generator(Postgres.Generator):
        TRANSFORMS = Postgres.Generator.TRANSFORMS.copy() | {
            sge.Interval: _interval,
            sge.GroupConcat: _group_concat,
            sge.ApproxDistinct: rename_func("approximate_count_distinct"),
        }
        TYPE_MAPPING = Postgres.Generator.TYPE_MAPPING.copy() | {
            sge.DataType.Type.TIMESTAMPTZ: "TIMESTAMP WITH LOCAL TIME ZONE",
        }


def _calculate_precision(interval_value: int) -> int:
    """Calculate interval precision.

    FlinkSQL interval data types use leading precision and fractional-
    seconds precision. Because the leading precision defaults to 2, we need to
    specify a different precision when the value exceeds 2 digits.

    (see
    https://learn.microsoft.com/en-us/sql/odbc/reference/appendixes/interval-literals)
    """
    # log10(interval_value) + 1 is equivalent to len(str(interval_value)), but is significantly
    # faster and more memory-efficient
    if interval_value == 0:
        return 0
    if interval_value < 0:
        raise ValueError(
            f"Expecting value to be a non-negative integer, got {interval_value}"
        )
    return int(math.log10(interval_value)) + 1


def _interval_with_precision(self, e):
    """Format interval with precision."""
    arg = e.args["this"].this
    formatted_arg = arg
    with contextlib.suppress(AttributeError):
        formatted_arg = arg.sql(self.dialect)

    unit = e.args["unit"]
    # when formatting interval scalars, need to quote arg and add precision
    if isinstance(arg, str):
        formatted_arg = f"'{formatted_arg}'"
        prec = _calculate_precision(int(arg))
        prec = max(prec, 2)
        unit.args["this"] += f"({prec})"

    return f"INTERVAL {formatted_arg} {unit}"


def _explode_to_unnest():
    """Convert explode into unnest.

    NOTE: Flink doesn't support UNNEST WITH ORDINALITY or UNNEST WITH OFFSET.
    """

    def _explode_to_unnest(expression: sge.Expression) -> sge.Expression:
        if isinstance(expression, sge.Select):
            from sqlglot.optimizer.scope import Scope

            taken_select_names = set(expression.named_selects)
            taken_source_names = {name for name, _ in Scope(expression).references}

            def new_name(names: set[str], name: str) -> str:
                name = find_new_name(names, name)
                names.add(name)
                return name

            # we use list here because expression.selects is mutated inside the loop
            for select in list(expression.selects):
                explode = select.find(sge.Explode)

                if explode:
                    explode_alias = ""

                    if isinstance(select, sge.Alias):
                        explode_alias = select.args["alias"]
                        alias = select
                    elif isinstance(select, sge.Aliases):
                        explode_alias = select.aliases[1]
                        alias = select.replace(sge.alias_(select.this, "", copy=False))
                    else:
                        alias = select.replace(sge.alias_(select, ""))
                        explode = alias.find(sge.Explode)
                        assert explode

                    explode_arg = explode.this

                    # This ensures that we won't use EXPLODE's argument as a new selection
                    if isinstance(explode_arg, sge.Column):
                        taken_select_names.add(explode_arg.output_name)

                    unnest_source_alias = new_name(taken_source_names, "_u")

                    if not explode_alias:
                        explode_alias = new_name(taken_select_names, "col")

                    alias.set("alias", sge.to_identifier(explode_alias))

                    column = sge.column(explode_alias, table=unnest_source_alias)

                    explode.replace(column)

                    expression.join(
                        sge.alias_(
                            sge.Unnest(
                                expressions=[explode_arg.copy()],
                            ),
                            unnest_source_alias,
                            table=[explode_alias],
                        ),
                        join_type="CROSS",
                        copy=False,
                    )

        return expression

    return _explode_to_unnest


class Flink(Hive):
    UNESCAPED_SEQUENCES = {"\\\\d": "\\d"}
    REGEXP_EXTRACT_DEFAULT_GROUP = 0

    class Generator(Hive.Generator):
        UNNEST_WITH_ORDINALITY = False

        TYPE_MAPPING = Hive.Generator.TYPE_MAPPING.copy() | {
            sge.DataType.Type.TIME: "TIME",
            sge.DataType.Type.STRUCT: "ROW",
        }

        TRANSFORMS = Hive.Generator.TRANSFORMS.copy() | {
            sge.Select: transforms.preprocess([_explode_to_unnest()]),
            sge.Stddev: rename_func("stddev_samp"),
            sge.StddevPop: rename_func("stddev_pop"),
            sge.StddevSamp: rename_func("stddev_samp"),
            sge.Variance: rename_func("var_samp"),
            sge.VariancePop: rename_func("var_pop"),
            sge.ArrayConcat: rename_func("array_concat"),
            sge.ArraySize: rename_func("cardinality"),
            sge.ArrayAgg: rename_func("array_agg"),
            sge.ArraySort: rename_func("array_sort"),
            sge.Length: rename_func("char_length"),
            sge.TryCast: lambda self, e: (
                f"TRY_CAST({e.this.sql(self.dialect)} AS {e.to.sql(self.dialect)})"
            ),
            sge.DayOfYear: rename_func("dayofyear"),
            sge.DayOfWeek: rename_func("dayofweek"),
            sge.DayOfMonth: rename_func("dayofmonth"),
            sge.Interval: _interval_with_precision,
        }

        # Flink is like Hive except where it might actually be convenient
        #
        # UNNEST works like the SQL standard, and not like Hive, so we have to
        # override sqlglot here and convince it that flink is not like Hive
        # when it comes to unnesting
        TRANSFORMS.pop(sge.Unnest, None)

        def unnest_sql(self, expression: sge.Unnest) -> str:
            return sgn.Generator.unnest_sql(self, expression)

        def struct_sql(self, expression: sge.Struct) -> str:
            from sqlglot.optimizer.annotate_types import annotate_types

            expression = annotate_types(expression)

            values = []
            schema = []

            for e in expression.expressions:
                if isinstance(e, sge.PropertyEQ):
                    e = sge.alias_(e.expression, e.this)
                # named structs
                if isinstance(e, sge.Alias):
                    if e.type and e.type.is_type(sge.DataType.Type.UNKNOWN):
                        self.unsupported(
                            "Cannot convert untyped key-value definitions (try annotate_types)."
                        )
                    else:
                        schema.append(f"{self.sql(e, 'alias')} {self.sql(e.type)}")
                    values.append(self.sql(e, "this"))
                else:
                    values.append(self.sql(e))

            if not (size := len(expression.expressions)) or len(schema) != size:
                return self.func("ROW", *values)
            return f"CAST(ROW({', '.join(values)}) AS ROW({', '.join(schema)}))"

        def array_sql(self, expression: sge.Array) -> str:
            # workaround for the time being because you cannot construct an array of named
            # STRUCTs directly from the ARRAY[] constructor
            # https://issues.apache.org/jira/browse/FLINK-34898
            from sqlglot.optimizer.annotate_types import annotate_types

            expression = annotate_types(expression)
            first_arg = seq_get(expression.expressions, 0)
            # it's an array of structs
            if isinstance(first_arg, sge.Struct):
                # get rid of aliasing because we want to compile this as CAST instead
                args = deepcopy(expression.expressions)
                for arg in args:
                    for e in arg.expressions:
                        arg.set("expressions", [e.unalias() for e in arg.expressions])

                format_values = ", ".join([self.sql(arg) for arg in args])
                # all elements of the array should have the same type
                format_dtypes = self.sql(first_arg.type)

                return f"CAST(ARRAY[{format_values}] AS ARRAY<{format_dtypes}>)"

            return (
                f"ARRAY[{', '.join(self.sql(arg) for arg in expression.expressions)}]"
            )

    class Tokenizer(Hive.Tokenizer):
        # In Flink, embedded single quotes are escaped like most other SQL
        # dialects: doubling up the single quote
        #
        # We override it here because we inherit from Hive's dialect and Hive
        # uses a backslash to escape single quotes
        STRING_ESCAPES = ["'"]


class Impala(Hive):
    NULL_ORDERING = "nulls_are_large"
    REGEXP_EXTRACT_DEFAULT_GROUP = 0

    class Generator(Hive.Generator):
        TRANSFORMS = Hive.Generator.TRANSFORMS.copy() | {
            sge.ApproxDistinct: rename_func("ndv"),
            sge.IsNan: rename_func("is_nan"),
            sge.IsInf: rename_func("is_inf"),
            sge.DayOfWeek: rename_func("dayofweek"),
            sge.Interval: lambda self, e: _interval(self, e, quote_arg=False),
            sge.CurrentDate: rename_func("current_date"),
        }


class MSSQL(TSQL):
    class Generator(TSQL.Generator):
        TRANSFORMS = TSQL.Generator.TRANSFORMS.copy() | {
            sge.ApproxDistinct: rename_func("approx_count_distinct"),
            sge.Stddev: rename_func("stdevp"),
            sge.StddevPop: rename_func("stdevp"),
            sge.StddevSamp: rename_func("stdev"),
            sge.Variance: rename_func("var"),
            sge.VariancePop: rename_func("varp"),
            sge.Ceil: rename_func("ceiling"),
            sge.DateFromParts: rename_func("datefromparts"),
        }


MySQL.Generator.TRANSFORMS |= {
    sge.LogicalOr: rename_func("max"),
    sge.LogicalAnd: rename_func("min"),
    sge.VariancePop: rename_func("var_pop"),
    sge.Variance: rename_func("var_samp"),
    sge.Stddev: rename_func("stddev_pop"),
    sge.StddevPop: rename_func("stddev_pop"),
    sge.StddevSamp: rename_func("stddev_samp"),
    sge.RegexpLike: (
        lambda _, e: f"({e.this.sql('mysql')} RLIKE {e.expression.sql('mysql')})"
    ),
}


def _create_sql(self, expression: sge.Create) -> str:
    properties = expression.args.get("properties")
    temporary = any(
        isinstance(prop, sge.TemporaryProperty)
        for prop in (properties.expressions if properties else [])
    )
    kind = expression.args["kind"]
    if kind.upper() in ("TABLE", "VIEW") and temporary:
        # Force insertion of required "GLOBAL" keyword
        expression_sql = self.create_sql(expression).replace(
            "CREATE TEMPORARY", "CREATE GLOBAL TEMPORARY"
        )
        if expression.expression:  #  CREATE ... AS ...
            return self.sql(expression_sql, "expression")
        else:  #  CREATE ... ON COMMIT PRESERVE ROWS
            # Autocommit does not work here for some reason so we append it manually
            return self.sql(
                expression_sql + " ON COMMIT PRESERVE ROWS",
                "expression",
            )
    return self.create_sql(expression)


# hack around https://github.com/tobymao/sqlglot/issues/3684
Oracle.NULL_ORDERING = "nulls_are_large"
Oracle.Generator.TRANSFORMS |= {
    sge.LogicalOr: rename_func("max"),
    sge.LogicalAnd: rename_func("min"),
    sge.VariancePop: rename_func("var_pop"),
    sge.Variance: rename_func("var_samp"),
    sge.Stddev: rename_func("stddev_pop"),
    sge.ApproxDistinct: rename_func("approx_count_distinct"),
    sge.Create: _create_sql,
    sge.Select: transforms.preprocess(
        [
            transforms.eliminate_semi_and_anti_joins,
            transforms.eliminate_distinct_on,
            transforms.eliminate_qualify,
        ]
    ),
    sge.GroupConcat: rename_func("listagg"),
}

# TODO: can delete this after bumping sqlglot version > 20.9.0
Oracle.Generator.TYPE_MAPPING |= {
    sge.DataType.Type.TIMETZ: "TIME",
    sge.DataType.Type.TIMESTAMPTZ: "TIMESTAMP",
}
Oracle.Generator.TZ_TO_WITH_TIME_ZONE = True


class Polars(Postgres):
    """Subclass of Postgres dialect for Polars.

    This is here to allow referring to the Postgres dialect as "polars"
    """


Postgres.Generator.TRANSFORMS |= {
    sge.Split: rename_func("string_to_array"),
    sge.RegexpSplit: rename_func("regexp_split_to_array"),
    sge.DateFromParts: rename_func("make_date"),
    sge.ArraySize: rename_func("cardinality"),
    sge.Pow: rename_func("pow"),
}


class PySpark(Spark):
    """Subclass of Spark dialect for PySpark.

    This is here to allow referring to the Spark dialect as "pyspark"
    """


class RisingWave(Postgres):
    # Need to disable timestamp precision
    # No "or replace" allowed in create statements
    # no "not null" clause for column constraints

    class Generator(Postgres.Generator):
        SINGLE_STRING_INTERVAL = True
        RENAME_TABLE_WITH_DB = False
        LOCKING_READS_SUPPORTED = True
        JOIN_HINTS = False
        TABLE_HINTS = False
        QUERY_HINTS = False
        NVL2_SUPPORTED = False
        PARAMETER_TOKEN = "$"  # noqa: S105
        TABLESAMPLE_SIZE_IS_ROWS = False
        TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
        SUPPORTS_SELECT_INTO = True
        JSON_TYPE_REQUIRED_FOR_EXTRACTION = True
        SUPPORTS_UNLOGGED_TABLES = True

        TYPE_MAPPING = Postgres.Generator.TYPE_MAPPING.copy() | {
            sge.DataType.Type.TIMESTAMPTZ: "TIMESTAMPTZ"
        }


Snowflake.Generator.TRANSFORMS |= {
    sge.ApproxDistinct: rename_func("approx_count_distinct"),
    sge.Levenshtein: rename_func("editdistance"),
}

SQLite.Generator.TYPE_MAPPING |= {sge.DataType.Type.BOOLEAN: "BOOLEAN"}


Trino.Generator.TRANSFORMS |= {
    sge.BitwiseLeftShift: rename_func("bitwise_left_shift"),
    sge.BitwiseRightShift: rename_func("bitwise_right_shift"),
    sge.FirstValue: rename_func("first_value"),
    sge.LastValue: rename_func("last_value"),
}



"""Tests for GizmoSQL vendored backend coverage.

Covers methods and branches in the vendored backend that are not
exercised by test_basic.py (single-backend, into_backend, caching).
"""

from __future__ import annotations

import contextlib
import json
import tempfile
from pathlib import Path
from urllib.parse import ParseResult

import pandas as pd
import pyarrow as pa
import pytest

import xorq.api as xo
import xorq.common.exceptions as exc


pytestmark = pytest.mark.gizmosql


# ── _Settings ───────────────────────────────────────────────────────────────


def test_settings_getitem(con):
    """_Settings.__getitem__ returns a setting value."""
    tz = con.settings["TimeZone"]
    assert isinstance(tz, str)
    assert tz == "UTC"


def test_settings_getitem_missing(con):
    """_Settings.__getitem__ raises KeyError for unknown settings."""
    with pytest.raises(KeyError):
        con.settings["__nonexistent_setting__"]


def test_settings_repr(con):
    """_Settings.__repr__ returns a non-empty string."""
    r = repr(con.settings)
    assert isinstance(r, str)
    assert len(r) > 0


# ── create_table edge cases ─────────────────────────────────────────────────


def test_create_table_schema_only(con, temp_table):
    """create_table with schema but no obj creates an empty table."""
    con.create_table(temp_table, schema={"a": "int64", "b": "string"})
    assert temp_table in con.list_tables()
    result = con.table(temp_table).execute()
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 0
    assert list(result.columns) == ["a", "b"]


def test_create_table_overwrite(con, temp_table):
    """create_table with overwrite=True replaces existing table."""
    t1 = xo.memtable({"x": [1, 2]})
    con.create_table(temp_table, t1)
    assert con.table(temp_table).count().execute() == 2

    t2 = xo.memtable({"x": [10, 20, 30]})
    con.create_table(temp_table, t2, overwrite=True)
    assert con.table(temp_table).count().execute() == 3


def test_create_table_from_dict(con, temp_table):
    """create_table with a dict (non-Expr obj) is converted via memtable."""
    con.create_table(temp_table, {"col_a": [1, 2, 3], "col_b": ["x", "y", "z"]})
    result = con.table(temp_table).execute()
    assert len(result) == 3


def test_create_table_no_obj_no_schema_raises(con, temp_table):
    """create_table without obj or schema raises ValueError."""
    with pytest.raises(ValueError, match="Either `obj` or `schema`"):
        con.create_table(temp_table)


def test_create_table_null_type_raises(con, temp_table):
    """create_table with a NULL typed column raises XorqTypeError."""
    with pytest.raises(exc.XorqTypeError, match="NULL typed columns"):
        con.create_table(temp_table, schema={"x": "null"})


# ── create_database / drop_database ─────────────────────────────────────────


def test_create_and_drop_database(con):
    """create_database and drop_database manage schemas."""
    db_name = "test_temp_schema"
    try:
        con.create_database(db_name)
        assert db_name in con.list_databases()
    finally:
        with contextlib.suppress(Exception):
            con.drop_database(db_name)


def test_create_database_with_catalog_raises(con):
    """create_database with a catalog argument raises."""
    with pytest.raises(exc.UnsupportedOperationError):
        con.create_database("test_db", catalog="some_catalog")


def test_drop_database_with_wrong_catalog_raises(con):
    """drop_database with a different catalog raises."""
    with pytest.raises(exc.UnsupportedOperationError):
        con.drop_database("test_db", catalog="nonexistent_catalog")


# ── list_databases with catalog filter ──────────────────────────────────────


def test_list_databases_with_catalog(con):
    """list_databases with catalog= filters results."""
    dbs = con.list_databases(catalog="memory")
    assert isinstance(dbs, list)
    assert "main" in dbs


# ── get_schema error handling ───────────────────────────────────────────────


def test_get_schema_table_not_found(con):
    """get_schema raises TableNotFound for nonexistent tables."""
    with pytest.raises(exc.TableNotFound):
        con.get_schema("__this_table_does_not_exist__")


# ── _convert_kwargs ─────────────────────────────────────────────────────────


def test_convert_kwargs():
    """_convert_kwargs converts read_only string to bool."""
    from xorq.vendor.ibis.backends.gizmosql import Backend

    kwargs = {"read_only": "true"}
    Backend._convert_kwargs(kwargs)
    assert kwargs["read_only"] is True

    kwargs = {"read_only": "False"}
    Backend._convert_kwargs(kwargs)
    assert kwargs["read_only"] is False


# ── read_csv ────────────────────────────────────────────────────────────────


def test_read_csv(con):
    """read_csv reads a local CSV file via _read_local_and_ingest."""
    with tempfile.NamedTemporaryFile(suffix=".csv", mode="w", delete=False) as f:
        f.write("id,name,value\n")
        f.write("1,alice,10.5\n")
        f.write("2,bob,20.3\n")
        f.write("3,charlie,30.1\n")
        csv_path = f.name

    try:
        t = con.read_csv(csv_path, table_name="test_read_csv")
        result = t.execute()
        assert isinstance(result, pd.DataFrame)
        assert len(result) == 3
        assert "name" in result.columns
    finally:
        with con._safe_raw_sql('DROP TABLE IF EXISTS "test_read_csv"'):
            pass
        Path(csv_path).unlink(missing_ok=True)


# ── read_json ───────────────────────────────────────────────────────────────


def test_read_json(con):
    """read_json reads newline-delimited JSON via _read_local_and_ingest."""
    with tempfile.NamedTemporaryFile(suffix=".json", mode="w", delete=False) as f:
        for row in [
            {"id": 1, "name": "alice"},
            {"id": 2, "name": "bob"},
            {"id": 3, "name": "charlie"},
        ]:
            f.write(json.dumps(row) + "\n")
        json_path = f.name

    try:
        t = con.read_json(json_path, table_name="test_read_json")
        result = t.execute()
        assert isinstance(result, pd.DataFrame)
        assert len(result) == 3
    finally:
        with con._safe_raw_sql('DROP TABLE IF EXISTS "test_read_json"'):
            pass
        Path(json_path).unlink(missing_ok=True)


# ── vendored execute (nested/null types) ────────────────────────────────────


def test_vendored_execute_with_nulls(con):
    """Vendored execute handles columns with null values."""
    from xorq.vendor.ibis.backends.gizmosql import Backend

    t = xo.memtable({"a": [1, None, 3], "b": ["x", None, "z"]})
    result = Backend.execute(con, t)
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 3


def test_vendored_execute_with_list_type(con):
    """Vendored execute handles nested (list) types via to_pylist()."""
    from xorq.vendor.ibis.backends.gizmosql import Backend

    with con._safe_raw_sql(
        "CREATE OR REPLACE TABLE test_nested AS SELECT [1, 2, 3] AS arr, 'a' AS label"
    ):
        pass
    try:
        t = con.table("test_nested")
        result = Backend.execute(con, t)
        assert isinstance(result, pd.DataFrame)
        assert len(result) == 1
    finally:
        with con._safe_raw_sql('DROP TABLE IF EXISTS "test_nested"'):
            pass


# ── vendored to_pyarrow (column / scalar) ──────────────────────────────────


def test_vendored_to_pyarrow_column(con, batting):
    """Vendored to_pyarrow returns ChunkedArray for column expressions."""
    from xorq.vendor.ibis.backends.gizmosql import Backend

    col_expr = batting.limit(5).playerID
    result = Backend.to_pyarrow(con, col_expr)
    assert isinstance(result, pa.ChunkedArray)
    assert len(result) == 5


def test_vendored_to_pyarrow_scalar(con, batting):
    """Vendored to_pyarrow returns a scalar for scalar expressions."""
    from xorq.vendor.ibis.backends.gizmosql import Backend

    scalar_expr = batting.count()
    result = Backend.to_pyarrow(con, scalar_expr)
    assert isinstance(result, pa.Scalar) or isinstance(result, int)


# ── vendored to_pyarrow_batches ─────────────────────────────────────────────


def test_vendored_to_pyarrow_batches(con, batting):
    """Vendored to_pyarrow_batches returns RecordBatchReader."""
    from xorq.vendor.ibis.backends.gizmosql import Backend

    reader = Backend.to_pyarrow_batches(con, batting.limit(10))
    assert isinstance(reader, pa.ipc.RecordBatchReader)
    table = reader.read_all()
    assert table.num_rows == 10


# ── _normalize_arrow_schema ─────────────────────────────────────────────────


def test_normalize_arrow_schema_large_types(con):
    """_normalize_arrow_schema downcasts large_string to string."""
    table = pa.table(
        {
            "a": pa.array(["x", "y", "z"], type=pa.large_string()),
            "b": pa.array([b"a", b"b", b"c"], type=pa.large_binary()),
            "c": pa.array([1, 2, 3], type=pa.int64()),
        }
    )
    result = con._normalize_arrow_schema(table)
    assert result.schema.field("a").type == pa.string()
    assert result.schema.field("b").type == pa.binary()
    assert result.schema.field("c").type == pa.int64()


def test_normalize_arrow_schema_no_change(con):
    """_normalize_arrow_schema returns table unchanged when no large types."""
    table = pa.table({"a": [1, 2, 3], "b": ["x", "y", "z"]})
    result = con._normalize_arrow_schema(table)
    assert result is table  # same object, no copy


# ── _register_in_memory_table (empty table) ─────────────────────────────────


def test_register_empty_memtable(con):
    """Empty memtable is created via DDL (ADBC ingest fails on zero rows)."""
    t = xo.memtable(
        pa.table(
            {"x": pa.array([], type=pa.int64()), "y": pa.array([], type=pa.string())}
        )
    )
    result = con.execute(t)
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 0


# ── _get_schema_using_query ─────────────────────────────────────────────────


def test_get_schema_using_query(con):
    """_get_schema_using_query returns Schema from DESCRIBE."""
    schema = con._get_schema_using_query("SELECT 1 AS a, 'hello' AS b")
    assert "a" in schema
    assert "b" in schema


# ── _from_url ───────────────────────────────────────────────────────────────


def test_from_url_parsing(con):
    """_from_url parses URL components correctly."""
    from xorq.vendor.ibis.backends.gizmosql import Backend

    url = ParseResult(
        scheme="gizmosql",
        netloc="myuser:mypass@myhost:31337",
        path="/mydb/myschema",
        params="",
        query="useEncryption=true&disableCertificateVerification=true",
        fragment="",
    )

    # We can't actually connect to this URL, but we can test the parsing
    # by mocking connect
    parsed_kwargs = {}

    original_connect = Backend.connect

    def mock_connect(self, **kwargs):
        parsed_kwargs.update(kwargs)
        raise ConnectionError("mock")

    Backend.connect = mock_connect
    try:
        backend = Backend()
        with pytest.raises(ConnectionError, match="mock"):
            backend._from_url(
                url, useEncryption="true", disableCertificateVerification="true"
            )
        assert parsed_kwargs["host"] == "myhost"
        assert parsed_kwargs["port"] == 31337
        assert parsed_kwargs["user"] == "myuser"
        assert parsed_kwargs["password"] == "mypass"
        assert parsed_kwargs["database"] == "mydb"
        assert parsed_kwargs["schema"] == "myschema"
        assert parsed_kwargs["use_encryption"] is True
        assert parsed_kwargs["disable_certificate_verification"] is True
    finally:
        Backend.connect = original_connect


# ── _create_temp_view / _get_temp_view_definition ───────────────────────────


def test_create_temp_view(con):
    """_create_temp_view creates a temporary view."""
    view_name = "test_temp_view_cov"
    try:
        con._create_temp_view(view_name, "SELECT 42 AS answer")
        with con._safe_raw_sql(f'SELECT * FROM "{view_name}"') as cur:
            result = cur.fetch_arrow_table()
        assert result.column("answer")[0].as_py() == 42
    finally:
        with contextlib.suppress(Exception):
            with con._safe_raw_sql(f'DROP VIEW IF EXISTS "{view_name}"'):
                pass


# ── from_connection ─────────────────────────────────────────────────────────


def test_from_connection(con):
    """from_connection creates a new backend from an existing ADBC connection."""
    from xorq.vendor.ibis.backends.gizmosql import Backend

    new_backend = Backend.from_connection(con.con)
    assert new_backend.current_database == "main"
    tables = new_backend.list_tables()
    assert isinstance(tables, list)



import pytest


TPC_H = [
    "tpc_h01",
    "tpc_h02",
    "tpc_h03",
    "tpc_h04",
    "tpc_h05",
    "tpc_h06",
    "tpc_h07",
    "tpc_h08",
    "tpc_h09",
    "tpc_h10",
    "tpc_h11",
    "tpc_h12",
    "tpc_h13",
    "tpc_h14",
    "tpc_h15",
    "tpc_h16",
    "tpc_h17",
    "tpc_h18",
    "tpc_h19",
    "tpc_h20",
    "tpc_h21",
    "tpc_h22",
]


@pytest.mark.parametrize("fixture_name", TPC_H)
def test_yaml_roundtrip(fixture_name, compiler, request):
    query = request.getfixturevalue(fixture_name)

    yaml_dict = compiler.to_yaml(query)
    roundtrip_query = compiler.from_yaml(yaml_dict)

    assert roundtrip_query.equals(query), (
        f"Roundtrip expression for {fixture_name} does not match the original."
    )



from xorq.vendor.ibis import selectors
from xorq.vendor.ibis.selectors import *  # noqa: F403
from xorq.vendor.ibis.selectors import index


__all__ = [  # noqa: PLE0604
    *selectors.__all__,
    "index",
]



from __future__ import annotations

import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis.backends.sql.compilers import PostgresCompiler
from xorq.vendor.ibis.backends.sql.compilers.base import ALL_OPERATIONS, NULL
from xorq.vendor.ibis.backends.sql.datatypes import RisingWaveType
from xorq.vendor.ibis.backends.sql.dialects import RisingWave


class RisingWaveCompiler(PostgresCompiler):
    __slots__ = ()

    dialect = RisingWave
    type_mapper = RisingWaveType

    UNSUPPORTED_OPS = (
        ops.Arbitrary,
        ops.Mode,
        ops.RandomUUID,
        ops.MultiQuantile,
        ops.ApproxMultiQuantile,
        *(
            op
            for op in ALL_OPERATIONS
            if issubclass(
                op, (ops.GeoSpatialUnOp, ops.GeoSpatialBinOp, ops.GeoUnaryUnion)
            )
        ),
    )

    def visit_DateNow(self, op):
        return self.cast(sge.CurrentTimestamp(), dt.date)

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the risingwave backend"
            )
        if not order_by:
            raise com.UnsupportedOperationError(
                "RisingWave requires an `order_by` be specified in `first`"
            )
        return self.agg.first_value(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the risingwave backend"
            )
        if not order_by:
            raise com.UnsupportedOperationError(
                "RisingWave requires an `order_by` be specified in `last`"
            )
        return self.agg.last_value(arg, where=where, order_by=order_by)

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                "RisingWave only implements `pop` correlation coefficient"
            )
        return super().visit_Correlation(
            op, left=left, right=right, how=how, where=where
        )

    def visit_Quantile(self, op, *, arg, quantile, where):
        if where is not None:
            arg = self.if_(where, arg, NULL)
        suffix = "cont" if op.arg.dtype.is_numeric() else "disc"
        return sge.WithinGroup(
            this=self.f[f"percentile_{suffix}"](quantile),
            expression=sge.Order(expressions=[sge.Ordered(this=arg)]),
        )

    visit_ApproxQuantile = visit_Quantile

    def visit_TimestampTruncate(self, op, *, arg, unit):
        unit_mapping = {
            "Y": "year",
            "Q": "quarter",
            "M": "month",
            "W": "week",
            "D": "day",
            "h": "hour",
            "m": "minute",
            "s": "second",
            "ms": "milliseconds",
            "us": "microseconds",
        }

        if (unit := unit_mapping.get(unit.short)) is None:
            raise com.UnsupportedOperationError(f"Unsupported truncate unit {unit}")

        return self.f.date_trunc(unit, arg)

    visit_TimeTruncate = visit_DateTruncate = visit_TimestampTruncate

    def _make_interval(self, arg, unit):
        return arg * sge.Interval(this=sge.convert(1), unit=self.v[unit.name])

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            return self.cast("".join(map(r"\x{:0>2x}".format, value)), dt.binary)
        elif dtype.is_date():
            return self.cast(value.isoformat(), dtype)
        elif dtype.is_json():
            return sge.convert(str(value))
        return None


compiler = RisingWaveCompiler()
